@@ -70,6 +70,7 @@ typedef unsigned short sampl_t;
 #define TRIG_RT		0x0008		/* perform op in real time */
 #define TRIG_CONFIG	0x0010		/* perform configuration, not triggering */
 #define TRIG_WAKE_EOS	0x0020		/* wake up on end-of-scan events */
+#define TRIG_WRITE	0x0040		/* write to bidirectional devices */
 
 /* subdevice flags */
 
@@ -155,7 +156,8 @@ struct comedi_trig_struct{
 	unsigned int trigsrc;
 	unsigned int trigvar;
 	unsigned int trigvar1;
-	unsigned int unused[4];
+	unsigned int data_len;
+	unsigned int unused[3];
 };
 
 struct comedi_chaninfo_struct{
@@ -224,17 +226,20 @@ struct comedi_devconfig_struct{
 
 /* callback stuff */
 
-#define COMEDI_CB_EOS		0	/* end of scan */
-#define COMEDI_CB_EOA		1	/* end of acquisition */
+#define COMEDI_CB_EOS		1	/* end of scan */
+#define COMEDI_CB_EOA		2	/* end of acquisition */
+#define COMEDI_CB_BLOCK		4	/* convenient block size */
+#define COMEDI_CB_EOBUF		8	/* end of buffer */
 
 /* exported functions */
 
 int comedi_trig_ioctl(unsigned int minor,unsigned int subdev,comedi_trig *it);
+int __comedi_trig_ioctl(unsigned int minor,unsigned int subdev,comedi_trig *it);
 int comedi_lock_ioctl(unsigned int minor,unsigned int subdev);
 int comedi_unlock_ioctl(unsigned int minor,unsigned int subdev);
 int comedi_cancel_ioctl(unsigned int minor,unsigned int subdev);
 int comedi_register_callback(unsigned int minor,unsigned int subdev,
-		unsigned int type,int (*cb)(void *),void *arg);
+		unsigned int mask,int (*cb)(unsigned int,void *),void *arg);
 
 #endif
 
