@@ -31,24 +31,26 @@
 #include <linux/malloc.h>
 #include <linux/errno.h>
 
-#if LINUX_VERSION_CODE < 0x020000
-#error kernel versions prior to 2.0 not supported
+
+#include <kern_compat.h>
+#ifdef CONFIG_COMEDI_RT
+#include <rt_printk.h>
+#include <rtl_compat.h>
 #else
-#if LINUX_VERSION_CODE < 0x020100
-#define LINUX_V20
-#else
-#define LINUX_V22
-#endif
+#define rt_printk(format,args...)	printk(format,##args)
+#define rt_printk_init()		
+#define rt_printk_cleanup()		
 #endif
 
 #include <comedi.h>
 
 #ifdef CONFIG_DEBUG
-#define DPRINTK(format, args...)	printk( format , ## args )
+#define DPRINTK(format, args...)	printk("comedi: " format , ## args )
 #else
 #define DPRINTK(format, args...)	/* */
 #endif
 
+
 typedef struct comedi_device_struct comedi_device;
 typedef struct comedi_subdevice_struct comedi_subdevice;
 
@@ -80,15 +82,17 @@ struct comedi_subdevice_struct{
 	int subdev_flags;
 	int timer_type;
 	int len_chanlist;		/* length of channel/gain list, if available */
-	unsigned int maxbufsz;
+
+	void		*prealloc_buf;		/* pre-allocated buffer */
+	unsigned int	prealloc_bufsz;		/* buffer size, in bytes */
 
 	void *lock;
 	void *busy;
 	
 	int io_bits;
 	
-	sampl_t maxdata;		/* if maxdata==0, use list */
-	sampl_t *maxdata_list;		/* list is channel specific */
+	lsampl_t maxdata;		/* if maxdata==0, use list */
+	lsampl_t *maxdata_list;		/* list is channel specific */
 	
 	unsigned int flags;
 	unsigned int *flaglist;
@@ -100,24 +104,23 @@ struct comedi_subdevice_struct{
 	
 	comedi_trig	cur_trig;	/* current trig structure */
 	
-	sampl_t		*buf;		/* pre-allocated buffer */
-
-	unsigned int buf_len;
 	volatile unsigned int buf_int_ptr;	/* buffer marker for interrupt */
-	unsigned int buf_user_ptr;	/* buffer marker for read() and write() */
+	unsigned int buf_user_ptr;		/* buffer marker for read() and write() */
+	volatile unsigned int buf_int_count;	/* byte count for interrupt */
+	unsigned int buf_user_count;		/* byte count for read() and write() */
 	unsigned int cur_chan;		/* useless channel marker for interrupt */
 	
+#if 0
 	unsigned int *range_list;	/* is this necessary? */
+#endif
 	
 	int (*trig[5])(comedi_device *,comedi_subdevice *,comedi_trig *);
 
 	int (*cancel)(comedi_device *,comedi_subdevice *);
 
-	int (*end_of_scan)(void *);
-	int (*end_of_acquisition)(void *);
-
-	void *eos_arg;
-	void *eoa_arg;
+	unsigned int cb_mask;
+	int (*cb_func)(unsigned int flags,void *);
+	void *cb_arg;
 
 	unsigned int state;
 };
@@ -140,7 +143,7 @@ struct comedi_device_struct{
 	int iosize;
 	int irq;
 
-	struct wait_queue *wait;
+	wait_queue_head_t wait;
 
 	/* still needed? */
 	comedi_buf rdbuf;
@@ -155,6 +158,10 @@ extern comedi_device *comedi_devices;
 extern comedi_devinit *comedi_devinits[];
 extern const int comedi_maxinits;
 
+#ifdef HAVE_RTL_POSIXIO
+extern struct rtl_file_operations comedi_rtl_fops;
+#endif
+
 /*
  * function prototypes
  */
@@ -162,6 +169,8 @@ extern const int comedi_maxinits;
 void comedi_error(comedi_device *dev,const char *s);
 void comedi_done(comedi_device *dev,comedi_subdevice *s);
 void comedi_eos(comedi_device *dev,comedi_subdevice *s);
+void comedi_eobuf(comedi_device *dev,comedi_subdevice *s);
+void comedi_bufcheck(comedi_device *dev,comedi_subdevice *s);
 
 void init_polling(void);
 void cleanup_polling(void);
@@ -171,6 +180,9 @@ void stop_polling(comedi_device *);
 void comedi_proc_init(void);
 void comedi_proc_cleanup(void);
 
+int di_unpack(unsigned int bits,comedi_trig *it);
+int do_pack(unsigned int *bits,comedi_trig *it);
+
 /*
    various internal comedi functions
  */
@@ -195,6 +207,7 @@ extern comedi_krange comedi_kranges[];
 #define TIMER_dt2814			2
 #define TIMER_atmio			3
 #define TIMER_acl8112			4
+#define TIMER_nanosec			5
 
 
 /* some silly little inline functions */
@@ -219,68 +232,11 @@ static inline int alloc_private(comedi_device *dev,int size)
 	return 0;
 }
 
-/* kernel compatibility */
-
-#ifdef LINUX_V22
-
-/* this is a useless abstraction that keeps several #ifdefs out of module.c */
-#define minor_of_file(x)        MINOR((x)->f_dentry->d_inode->i_rdev)
-
-#endif
 
 #ifdef LINUX_V20
-
-#define minor_of_file(x)        MINOR((x)->f_inode->i_rdev)
-
-#define ioremap vremap
-#define iounmap vfree
-
-#define signal_pending(x)	(((x)->signal) & (~(x)->blocked))
-
-#if CONFIG_EXPORT
+#ifdef CONFIG_EXPORT
 extern struct symbol_table comedi_syms;
 #endif
-
-#include <asm/segment.h>
-
-static inline int copy_to_user(void * to,const void *from,unsigned long n_bytes)
-{
-	int i;
-
-	if((i=verify_area(VERIFY_WRITE,to,n_bytes)) != 0)
-		return i;
-	memcpy_tofs(to,from,n_bytes);
-	return 0;
-}
-
-static inline int copy_from_user(void * to,const void *from,unsigned long n_bytes)
-{
-	int i;
-	if((i=verify_area(VERIFY_READ,from,n_bytes))!=0)
-		return i;
-	memcpy_fromfs(to,from,n_bytes);
-	return 0;
-}
-
-static inline int clear_user(void * mem,unsigned long len)
-{
-	char *cmem=mem;
-	
-	if(verify_area(VERIFY_WRITE,mem,len))
-		return len;
-	/* this is slow, but I'm lazy */
-	while(len--){put_user(0,cmem);cmem++;}
-	return 0;
-}
-
-#endif
-
-
-#ifdef LINUX_V20
-/* actually LINUX_VERSION_CODE < 0x020155 */
-
-#define LINUX_old_PCI_compatibility
-
 #endif
 
 
