@@ -135,71 +135,85 @@ static int range_types[] =
 
 static int das1600_ai(comedi_device * dev, comedi_subdevice * s, comedi_trig * it)
 {
-	int i, gain, hi, lo, chan;
-#if 0
-	unsigned long flags;
-#endif
+  int i;
+  for(i=0 ; i < it->n_chan ; i++) {
+    int t, gain, hi, lo, chan;
 
 	if (it->mode != 0)
 		return -EINVAL;
 
-	gain = CR_RANGE(it->chanlist[0]);
-	chan = CR_CHAN(it->chanlist[0]);
-#if 0
-	save_flags(flags);
-	cli();
-#endif
+    gain = CR_RANGE(it->chanlist[i]);
+    chan = CR_CHAN(it->chanlist[i]);
 
 	outb(gain, dev->iobase + DAS1600_Gain_Control);
 	outb(chan + (chan << 4), dev->iobase + DAS1600_Channel_Mux);
 	outb(0, dev->iobase + DAS1600_ADC_Low);
-	for (i = 0; i < DAS1600_TIMEOUT; i++) {
+    for (t = 0; t < DAS1600_TIMEOUT; t++) {
 		if ((inb(dev->iobase + DAS1600_Status) & 0x80) == 0) {
 			break;
 		}
 	}
-	if (i == DAS1600_TIMEOUT) {
-		printk("das1600: timeout\n");
+    if (t == DAS1600_TIMEOUT) {
+		rt_printk("das1600: timeout\n");
 	}
 	lo = inb(dev->iobase + DAS1600_ADC_Low);
 	hi = inb(dev->iobase + DAS1600_ADC_High);
 
-#if 0
-	restore_flags(flags);
-#endif
 	switch (devpriv->card) {
 	case card_1601_12:
 	case card_1602_12:
-		it->data[0] = (hi << 4) | (lo >> 4);
+       it->data[i] = (hi << 4) | (lo >> 4);
 		break;
 	case card_1602_16:
-		it->data[0] = (hi << 8) | lo;
+       it->data[i] = (hi << 8) | lo;
 		break;
 	}
-
-	return 0;
+  }
+  return i;
 }
 
 
 static int das1600_ao(comedi_device * dev, comedi_subdevice * s, comedi_trig * it)
 {
+  int i;
+  for(i=0 ; i < it->n_chan ; i++) {
 	int chan;
 	int data;
-	int i;
 
-	for(i=0;i<it->n_chan;i++){
 		chan = CR_CHAN(it->chanlist[i]);
 		data = it->data[i];
 
-		outb((data & 0x00f) << 4, dev->iobase + (chan) ? DAS1600_DA1_Low : DAS1600_DA0_Low);
-		outb((data & 0xff0) >> 4, dev->iobase + (chan) ? DAS1600_DA1_High : DAS1600_DA0_High);
+    outb((data & 0x00f) << 4, dev->iobase + ((chan) ? DAS1600_DA1_Low : DAS1600_DA0_Low));
+    outb((data & 0xff0) >> 4, dev->iobase + ((chan) ? DAS1600_DA1_High : DAS1600_DA0_High));
 	}
-
 	return i;
 }
 
-#if 0
-/* These don't appear to be correct */
+#if 1
+
+static int das1600_di(comedi_device * dev, comedi_subdevice * s, comedi_trig * it)
+{
+	unsigned int bits;
+
+	bits = inb(dev->iobase + DAS1600_Digital_4_Bit);
+
+	di_unpack(bits,it);
+
+	return it->n_chan;
+}
+
+static int das1600_do(comedi_device * dev, comedi_subdevice * s, comedi_trig * it)
+{
+	do_pack(&s->state,it);
+
+       /* the outputs are inverted... dumb... (Is this really true? can't
+          find it in my docs...) */
+	outb(s->state ^ 0xff, dev->iobase + DAS1600_Digital_4_Bit);
+
+	return it->n_chan;
+}
+
+#else
 
 static int das1600_di(comedi_device * dev, comedi_subdevice * s, comedi_trig * it)
 {
@@ -239,15 +253,47 @@ static int das1600_do(comedi_device * dev, comedi_subdevice * s, comedi_trig * i
 
 	return i;
 }
+
 #endif
 
 
 /*
-   options[0]   i/o base
-   [1]  
-   [2]  
-   [3]  
-   [4]  ai single-ended/differential
+  options[0]   Board base address
+  options[1]   IRQ
+  options[2]   DMA level select configuration
+                 0 == DMA level 1
+                 1 == DMA level 1
+                 3 == DMA level 3
+  options[3]   Crystal select configuration
+                 0 == 1 MHz
+                 1 == 1 MHz
+                 10 == 10 MHz
+  options[4]   Input configuration
+                 0 == differential
+                 1 == single-ended
+  options[5]   Analog input range configuration
+                 0 == bipolar 10V  (-10V -- +10V)
+                 1 == unipolar 10V  (0V -- +10V)
+  options[6]   Analog output 0 range configuration
+                 0 == bipolar 10V  (-10V -- +10V)
+                 1 == bipolar 5V  (-5V -- +5V)
+                 2 == bipolar user supplied  (-xV -- +xV)
+                 3 == unipolar 10V  (0V -- +10V)
+                 4 == unipolar 5V  (0V -- +5V)
+                 5 == unipolar user supplied  (0V -- +xV)
+  options[7]   Analog output 1 range configuration
+                 0 == bipolar 10V  (-10V -- +10V)
+                 1 == bipolar 5V  (-5V -- +5V)
+                 2 == bipolar user supplied  (-xV -- +xV)
+                 3 == unipolar 10V  (0V -- +10V)
+                 4 == unipolar 5V  (0V -- +5V)
+                 5 == unipolar user supplied  (0V -- +xV)
+  options[8]   8255 I/O bits direction (0 == input, 1 == output)
+                 0x0000bb == port A
+                 0x00bb00 == port B
+                 0x0b0000 == port C low
+                 0xb00000 == port C high
+
  */
 int das1600_init(comedi_device * dev, comedi_devconfig * it)
 {
@@ -287,15 +333,10 @@ int das1600_init(comedi_device * dev, comedi_devconfig * it)
 
 	dev->n_subdevices=5;
 
-	dev->private = kmalloc(sizeof(das1600_private), GFP_KERNEL);
-	dev->subdevices = kmalloc(dev->n_subdevices * sizeof(comedi_subdevice), GFP_KERNEL);
-
-	if (!dev->private || !dev->subdevices) {
-		result = -ENOMEM;
+	if((result=alloc_private(dev,sizeof(das1600_private)))<0)
+		goto cleanup;
+	if((result=alloc_subdevices(dev))<0)
 		goto cleanup;
-	}
-	memset(dev->private, 0, sizeof(das1600_private));
-	memset(dev->subdevices, 0, dev->n_subdevices * sizeof(comedi_subdevice));
 
 	devpriv->card = board;
 	devpriv->dma = it->options[2];
@@ -350,32 +391,25 @@ int das1600_init(comedi_device * dev, comedi_devconfig * it)
 
 	s++;
 	/* di subdevice */
-#if 0
 	s->type = COMEDI_SUBD_DI;
 	s->subdev_flags = SDF_READABLE;
 	s->trig[0] = das1600_di;
-	s->n_chan = 8;
+       s->n_chan = 4;
 	s->maxdata = 1;
 	s->range_type = RANGE_digital;
-#else
-	s->type = COMEDI_SUBD_UNUSED;
-#endif
 
 	s++;
-#if 0
 	/* do subdevice */
 	s->type = COMEDI_SUBD_DO;
 	s->subdev_flags = SDF_WRITEABLE;
 	s->trig[0] = das1600_do;
-	s->n_chan = 8;
+       s->n_chan = 4;
 	s->maxdata = 1;
 	s->range_type = RANGE_digital;
-#else
-	s->type = COMEDI_SUBD_UNUSED;
-#endif
 
 	s++;
 	/* 8255 subdevice */
+       s->io_bits = it->options[8];
 	subdev_8255_init(dev,s,NULL,(void *)(dev->iobase+DAS1600_Port_A));
 
 	dev->rem = das1600_rem;
@@ -389,12 +423,6 @@ int das1600_init(comedi_device * dev, comedi_devconfig * it)
 
 static void das1600_release_resources(comedi_device * dev)
 {
-	if (dev->private) {
-		kfree(dev->private);
-	}
-	if (dev->subdevices) {
-		kfree(dev->subdevices);
-	}
 	if (dev->iobase) {
 		release_region(dev->iobase, DAS1600_BASE_SIZE);
 		release_region(dev->iobase + 0x400, DAS1600_DIO_SIZE);
