@@ -189,9 +189,6 @@ static int dt2814_rem(comedi_device *dev)
 {
 	printk("comedi%d: dt2814: remove\n",dev->minor);
 	
-	kfree(dev->private);
-	kfree(dev->subdevices);
-
 	if(dev->irq){
 		free_irq(dev->irq,dev);
 	}
