@@ -36,7 +36,12 @@
 #include <asm/dma.h>
 #include <comedi_module.h>
 
-#define DT_DMA 0
+/*
+*    DMA is corrupting memory.  disabled.
+*/
+#define DT_DMA 1
+
+#define DEBUG
 
 #define DT2821_TIMEOUT		100	/* 500 us */
 #define DT2821_SIZE 0x10
@@ -59,6 +64,30 @@
  *  masks and values are as follows:
  */
 
+#define RELAXED_FINGERPRINT
+
+#ifdef RELAXED_FINGERPRINT
+
+#define DT2821_ADCSR_MASK 0xfff0
+#define DT2821_ADCSR_VAL 0x7c00
+
+#define DT2821_CHANCSR_MASK 0xf0f0
+#define DT2821_CHANCSR_VAL 0x70f0
+
+#define DT2821_DACSR_MASK 0x7c93
+#define DT2821_DACSR_VAL 0x7c90
+
+#define DT2821_SUPCSR_MASK 0xf8ff
+#define DT2821_SUPCSR_VAL 0x0000
+
+#define DT2821_TMRCTR_MASK 0xff00
+#define DT2821_TMRCTR_VAL 0xf000
+
+#else
+
+/* this is the fingerprint list for the
+   pre-DT-ez boards */
+
 #define DT2821_ADCSR_MASK 0xfff0
 #define DT2821_ADCSR_VAL 0x7c00
 
@@ -74,6 +103,8 @@
 #define DT2821_TMRCTR_MASK 0xff00
 #define DT2821_TMRCTR_VAL 0xf000
 
+#endif
+
 /*
  *    Bit fields of each register
  */
@@ -134,7 +165,7 @@ RANGE_dt282x_ai_lo_bipolar
         -5      5
         -2.5    2.5
         -1.25   1.25
-RANGE_dt282x_ai_lo_bipolar
+RANGE_dt282x_ai_lo_unipolar
         0       10
         0       5
         0       2.5
@@ -159,69 +190,114 @@ RANGE_dt282x_ai_hi_unipolar
         0       1
         0       0.1
         0       0.02
+RANGE_bipolar2_5
+	-2.5	2.5
 ---END-RANGE-DEFS---
 */
 
-
-/*
- *    map of boardnames and single ended/differential jumper
- *      to boardtype below
- */
 typedef struct {
 	char *name;
-	int boardcode_se;
-	int boardcode_di;
-	int t0, t1, t2;
-} boardname_t;
-
-
-static boardname_t boardnames[] =
-{
-	{"dt2821", 0, 1, RANGE_dt282x_ai_lo_unipolar, RANGE_dt282x_ai_lo_bipolar, 0},
-	{"dt2821-f-16se", 0, 0, RANGE_dt282x_ai_lo_unipolar, RANGE_dt282x_ai_lo_bipolar, 0},
-	{"dt2821-f-8di", 1, 1, RANGE_dt282x_ai_lo_unipolar, RANGE_dt282x_ai_lo_bipolar, 0},
-	{"dt2821-g-16se", 0, 0, RANGE_dt282x_ai_lo_unipolar, RANGE_dt282x_ai_lo_bipolar, RANGE_dt282x_ai_5_bipolar},
-	{"dt2821-g-8di", 1, 1, RANGE_dt282x_ai_lo_unipolar, RANGE_dt282x_ai_lo_bipolar, RANGE_dt282x_ai_5_bipolar},
-	{"dt2823", 2, 2, (RANGE_bipolar10), RANGE_bipolar10, 0},
-	{"dt2824-pgh", 3, 4, RANGE_dt282x_ai_lo_unipolar, RANGE_dt282x_ai_lo_bipolar, 0},
-	{"dt2824-pgl", 3, 4, RANGE_dt282x_ai_hi_unipolar, RANGE_dt282x_ai_hi_bipolar, 0},
-	{"dt2825", 0, 1, RANGE_dt282x_ai_hi_unipolar, RANGE_dt282x_ai_hi_bipolar, 0},
-	{"dt2827", 5, 5, (RANGE_bipolar10), RANGE_bipolar10, 0},
-	{"dt2828", 6, 6, RANGE_unipolar10, RANGE_bipolar10, 0}
-};
-static int n_boardnames = 11;
-
-/*
- *    These boardtypes represent the different possible combinations
- *      of boards that exist.  This includes the A/D input jumper
- *      determining if the inputs are single ended or differential.
- *      Other jumpers are not important.
- */
-/* beware, the macros from hell */
-#define _12BIT	12, 0xfff, 0x800
-#define _16BIT	16, 0xffff, 0x8000
-typedef struct {
 	int adbits;
-	int admask;
-	int adsignbit;
-	int adgains;
-	int adchans;
-	int adref;
+	int adchan_se;
+	int adchan_di;
+	int ispgl;
+	int dachan;
 	int dabits;
-	int damask;
-	int dasignbit;
-	int dachans;
 } boardtype_t;
+
 static boardtype_t boardtypes[] =
 {
-	{_12BIT, 4, 16, AREF_COMMON, _12BIT, 2},	/* 2821, 2825, single ended */
-	{_12BIT, 4, 8, AREF_DIFF, _12BIT, 2},	/* 2821, 2825, differential */
-	{_16BIT, 1, 4, AREF_DIFF, _16BIT, 2},	/* 2823 */
-	{_12BIT, 4, 16, AREF_COMMON, 0, 0, 0, 0},	/* 2824, single ended */
-	{_12BIT, 4, 8, AREF_DIFF, 0, 0, 0, 0},	/* 2824, differential */
-	{_16BIT, 1, 4, AREF_DIFF, _12BIT, 2},	/* 2827 */
-	{_12BIT, 1, 4, AREF_COMMON, _12BIT, 2},	/* 2828 */
+	{	name:		"dt2821",
+		adbits:		12,
+		adchan_se:	16,
+		adchan_di:	8,
+		ispgl:		0,
+		dachan:		2,
+		dabits:		12,
+	},
+	{	name:		"dt2823",
+		adbits:		16,
+		adchan_se:	0,
+		adchan_di:	4,
+		ispgl:		0,
+		dachan:		2,
+		dabits:		16,
+	},
+	{	name:		"dt2824-pgh",
+		adbits:		16,
+		adchan_se:	16,
+		adchan_di:	8,
+		ispgl:		0,
+		dachan:		0,
+		dabits:		0,
+	},
+	{	name:		"dt2824-pgl",
+		adbits:		16,
+		adchan_se:	16,
+		adchan_di:	8,
+		ispgl:		1,
+		dachan:		0,
+		dabits:		0,
+	},
+	{	name:		"dt2825",
+		adbits:		12,
+		adchan_se:	16,
+		adchan_di:	8,
+		ispgl:		0,
+		dachan:		2,
+		dabits:		12,
+	},
+	{	name:		"dt2827",
+		adbits:		16,
+		adchan_se:	0,
+		adchan_di:	4,
+		ispgl:		0,
+		dachan:		2,
+		dabits:		12,
+	},
+	{	name:		"dt2828",
+		adbits:		12,
+		adchan_se:	4,
+		adchan_di:	0,
+		ispgl:		0,
+		dachan:		2,
+		dabits:		12,
+	},
+	{	name:		"dt21-ez",
+		adbits:		12,
+		adchan_se:	16,
+		adchan_di:	8,
+		ispgl:		0,
+		dachan:		2,
+		dabits:		12,
+	},
+	{	name:		"dt23-ez",
+		adbits:		16,
+		adchan_se:	16,
+		adchan_di:	8,
+		ispgl:		0,
+		dachan:		0,
+		dabits:		0,
+	},
+	{	name:		"dt24-ez",
+		adbits:		12,
+		adchan_se:	16,
+		adchan_di:	8,
+		ispgl:		0,
+		dachan:		0,
+		dabits:		0,
+	},
+	{	name:		"dt24-ez-pgl",
+		adbits:		12,
+		adchan_se:	16,
+		adchan_di:	8,
+		ispgl:		1,
+		dachan:		0,
+		dabits:		0,
+	},
 };
+static int n_boardtypes = sizeof(boardtypes)/sizeof(boardtype_t);
+
 
 typedef struct {
 	boardtype_t *board;	/* type of board we have        */
@@ -230,12 +306,14 @@ typedef struct {
 	int da0_2scomp;		/* same, for DAC0               */
 	int da1_2scomp;		/* same, for DAC1               */
 
+	int darangelist[2];
+
 	int dacsr;		/* software copies of registers */
 	int adcsr;
 	int supcsr;
 
 	int ntrig;
-	int curadchan;
+	int nread;
 
 	short *dmaptr1, *dmaptr2;	/* pointers to DMA buffers      */
 	int dma1, dma2;		/* DMA channels                 */
@@ -276,110 +354,135 @@ typedef struct {
 comedi_devinit dt282x_init;
 static int dt282x_rem(comedi_device * dev);
 static void free_resources(comedi_device *dev);
+static int prep_dma(comedi_device * dev);
 
 
 #if DT_DMA
+static int dt282x_grab_dma(comedi_device *dev,int dma1,int dma2);
+
 /*
  *    We don't really care here if we're doing dual DMA or not,
  *      since we just restart the DMA controller.
  */
-static void dt282x_dma_interrupt(comedi_device * dev)
+static void dt282x_dma_interrupt(comedi_device * dev,comedi_subdevice *s)
 {
-	int i, n, dma_chan;
-	comedi_sample pt;
-
-	pt.chan = devpriv->curadchan;
-	pt.job = dev->busy[0];
+	int i, data;
 
+#if 0
+printk("dma interrupt\n");
+#endif
 	update_supcsr(DT2821_CLRDMADNE);
 
 	if (!devpriv->current_dma_chan) {
-		printk("intr dma 0\n");
+#if 0
+		rt_printk("intr dma 0\n");
+#endif
 		/* channel that finished was A */
 		disable_dma(devpriv->dma1);
 		for (i = 0; i < devpriv->dma1_size; i++) {
-			pt.data = devpriv->dmaptr1[i];
-			report_sample(dev, &pt);
+			data=devpriv->dmaptr1[i];
+			data&=(1<<boardtype.adbits)-1;
+			if(devpriv->ad_2scomp){
+				data^=1<<(boardtype.adbits-1);
+			}
+			s->cur_trig.data[s->buf_int_ptr++] = data;
 		}
+		devpriv->nread-=devpriv->dma1_size;
+		devpriv->current_dma_chan=1;
 	} else {
-		printk("intr dma 1\n");
+#if 0
+		rt_printk("intr dma 1\n");
+#endif
 		/* channel that finished was B */
 		disable_dma(devpriv->dma2);
 		for (i = 0; i < devpriv->dma2_size; i++) {
-			pt.data = devpriv->dmaptr2[i];
-			report_sample(dev, &pt);
+			data=devpriv->dmaptr2[i];
+			data&=(1<<boardtype.adbits)-1;
+			if(devpriv->ad_2scomp){
+				data^=1<<(boardtype.adbits-1);
+			}
+			s->cur_trig.data[s->buf_int_ptr++] = data;
 		}
+		devpriv->nread-=devpriv->dma2_size;
+		devpriv->current_dma_chan=0;
 	}
-	if (!devpriv->ntrig) {
-		/* we're done.  let's go home. */
+	if(devpriv->nread<0){
+		printk("dt282x: off by one\n");
+		devpriv->nread=0;
+	}
+	if (!devpriv->nread) {
+		devpriv->adcsr=0;
+		update_adcsr(0);
 
-		dev->busy[0] = 0;
-		comedi_done(dev,dev->subdevices+0);
+		/* this eliminates "extra sample" issues */
+		devpriv->supcsr = 0;
+		update_supcsr(DT2821_ADCINIT);
+
+		comedi_done(dev,s);
 
 		return;
 	}
-	/* restart the channel */
-	n = devpriv->ntrig;
-	if (n >= devpriv->dma_maxsize) {
-		n = devpriv->dma_maxsize;
-	} else {
-		/* clear the dual dma flag, making this the last dma segment */
+	if (!devpriv->ntrig)
+		return;
+
+#if 1
+	/* clear the dual dma flag, making this the last dma segment */
+	/* XXX probably wrong */
+	if(!devpriv->ntrig){
 		devpriv->supcsr &= ~(DT2821_DDMA);
 		update_supcsr(0);
 	}
-	devpriv->ntrig -= n;
+#endif
+	/* restart the channel */
+	prep_dma(dev);
 
-	if (!devpriv->current_dma_chan) {
-		/* channel A */
-		devpriv->dma1_size = n;
-		dma_chan = devpriv->dma1;
-/* since dma functions are all inlined, two copies are better than one */
-		{
-			long flags;
-			save_flags(flags);
-			cli();
-			clear_dma_ff(5);	/* actual channel doesn't matter */
-			set_dma_count(dma_chan, n << 1);
-			enable_dma(dma_chan);
-			restore_flags(flags);
-		}
-		devpriv->current_dma_chan = 1;
-	} else {
-		/* channel B */
-		devpriv->dma2_size = n;
-		dma_chan = devpriv->dma2;
-		{
-			long flags;
-			save_flags(flags);
-			cli();
-			clear_dma_ff(5);	/* actual channel doesn't matter */
-			set_dma_count(dma_chan, n << 1);
-			enable_dma(dma_chan);
-			restore_flags(flags);
-		}
-		devpriv->current_dma_chan = 0;
-	}
 
 	return;
 }
 #endif
 
-#if 0
-/*
- *    This function is incomplete
- */
-static int setup_dma_chan(comedi_device * dev, int chan, int count)
+static int prep_dma(comedi_device * dev)
 {
 	int n;
-	long flags;
+	unsigned long flags;
+	int dma_chan;
+	unsigned long dma_ptr;
 
-	n = count;
-	if (n >= dev->dma_maxsize)
+	if(!devpriv->ntrig)
+		return 0;
+
+	n = devpriv->ntrig;
+	if (n >= devpriv->dma_maxsize)
 		n = devpriv->dma_maxsize;
-	dev->ntrig -= n;
-	/* ... */
-}
+	devpriv->ntrig -= n;
+
+	/*
+	   current_dma_chan is the one that is currently
+	   _running_, not the one we are programming here
+	*/
+	if (devpriv->current_dma_chan) {
+		devpriv->dma1_size = n;
+		dma_chan = devpriv->dma1;
+		dma_ptr = virt_to_bus(devpriv->dmaptr1);
+	}else{
+		devpriv->dma2_size = n;
+		dma_chan = devpriv->dma2;
+		dma_ptr = virt_to_bus(devpriv->dmaptr2);
+		set_dma_count(dma_chan, n << 1);
+	}
+#if 0
+printk("prepping dma %d for %d samples\n",dma_chan,n);
 #endif
+	flags=claim_dma_lock();
+	clear_dma_ff(5);	/* actual channel doesn't matter */
+	set_dma_mode(dma_chan, DMA_MODE_READ);
+	set_dma_addr(dma_chan, dma_ptr);
+	set_dma_count(dma_chan, n << 1);
+	enable_dma(dma_chan);
+	release_dma_lock(flags);
+
+	return n;
+}
 
 static void dt282x_interrupt(int irq, void *d, struct pt_regs *regs)
 {
@@ -388,19 +491,18 @@ static void dt282x_interrupt(int irq, void *d, struct pt_regs *regs)
 	unsigned int supcsr, adcsr, dacsr;
 	sampl_t data;
 
-	/*
-	   XXX these should be rearranged so that the most probable
-	   case is near the top.
-	 */
+/*printk("interrupt\n");*/
+	supcsr = inw(dev->iobase + DT2821_SUPCSR);
+/*printk("supcsr=%02x\n",supcsr);*/
 #if DT_DMA
-	if ((supcsr = inw(dev->iobase + DT2821_SUPCSR)) & DT2821_DMAD) {
-		dt282x_dma_interrupt(dev);
+	if (supcsr & DT2821_DMAD) {
+		dt282x_dma_interrupt(dev,s);
 		return;
 	}
-#else
-	supcsr=0;	/* annoying warnings... */
 #endif
-	if ((adcsr = inw(dev->iobase + DT2821_ADCSR)) & DT2821_ADERR) {
+	adcsr=inw(dev->iobase + DT2821_ADCSR);
+/*printk("adcsr=%02x\n",adcsr);*/
+	if (adcsr & DT2821_ADERR) {
 		comedi_error(dev, "A/D error");
 		return;
 	}
@@ -410,12 +512,20 @@ static void dt282x_interrupt(int irq, void *d, struct pt_regs *regs)
 	}
 	if (adcsr & DT2821_ADDONE) {
 		data = (sampl_t) inw(dev->iobase + DT2821_ADDAT);
-
-		if (--devpriv->ntrig) {
-			update_supcsr(DT2821_STRIG);
-		} else {
-			comedi_done(dev,s);
+		data&=(1<<boardtype.adbits)-1;
+		if(devpriv->ad_2scomp){
+			data^=1<<(boardtype.adbits-1);
 		}
+		s->cur_trig.data[s->buf_int_ptr++]=data;
+
+		devpriv->nread--;
+		if(!devpriv->nread){
+			comedi_done(dev,s);
+		}else{
+			if(supcsr&DT2821_SCDN)
+				update_supcsr(DT2821_STRIG);
+		}
+
 		return;
 	}
 }
@@ -461,7 +571,7 @@ static int dt282x_ai_mode0(comedi_device * dev, comedi_subdevice * s, comedi_tri
 		 return -ETIME;
 	    );
 
-	it->data[0] = inw(dev->iobase + DT2821_ADDAT) & boardtype.admask;
+	it->data[0] = inw(dev->iobase + DT2821_ADDAT) & ((1<<boardtype.adbits)-1);
 	if (devpriv->ad_2scomp)
 		it->data[0] ^= (1 << (boardtype.adbits - 1));
 
@@ -470,14 +580,19 @@ static int dt282x_ai_mode0(comedi_device * dev, comedi_subdevice * s, comedi_tri
 
 static int dt282x_ai_mode1(comedi_device * dev, comedi_subdevice * s, comedi_trig * it)
 {
-	if (DT_DMA && !devpriv->usedma) {
+	if (!DT_DMA || !devpriv->usedma) {
 		dt282x_load_changain(dev,it->n_chan,it->chanlist);
 
+		devpriv->ntrig=it->n*it->n_chan;
+		devpriv->nread=devpriv->ntrig;
+
 		outw(it->trigvar, dev->iobase + DT2821_TMRCTR);
 
 		devpriv->adcsr = DT2821_ADCLK | DT2821_IADDONE;
 		update_adcsr(0);
 
+		devpriv->supcsr = DT2821_ERRINTEN ;
+
 		update_supcsr(DT2821_PRLD);
 		wait_for(!mux_busy(),
 			 comedi_error(dev, "timeout\n");
@@ -488,64 +603,24 @@ static int dt282x_ai_mode1(comedi_device * dev, comedi_subdevice * s, comedi_tri
 		return 0;
 	} else {
 #if DT_DMA
-		int n;
-
-		if (it->n > 2048)
-			return -EINVAL;
 		outw(it->trigvar, dev->iobase + DT2821_TMRCTR);
 
 		devpriv->supcsr = DT2821_ERRINTEN | DT2821_DS0;
 		update_supcsr(DT2821_CLRDMADNE | DT2821_BUFFB | DT2821_ADCINIT);
 		devpriv->adcsr = 0;
 
-		n = devpriv->ntrig;
-		if (n > devpriv->dma_maxsize) {
-			n = devpriv->dma_maxsize;
+		devpriv->ntrig=it->n*it->n_chan;
+		devpriv->nread=devpriv->ntrig;
+
+		devpriv->current_dma_chan=1;
+		prep_dma(dev);
+		devpriv->current_dma_chan=0;
+		if(devpriv->ntrig){
+			prep_dma(dev);
 			devpriv->supcsr |= DT2821_DDMA;
 			update_supcsr(0);
-		} else {
-			/* clear the dual dma flag, making this the only dma segment */
-			devpriv->supcsr &= ~(DT2821_DDMA);
-			update_supcsr(0);
 		}
-		devpriv->ntrig -= n;
-		devpriv->dma1_size = n;
-/* DMA junk */
-		{
-			long flags;
-			save_flags(flags);
-			cli();
-			clear_dma_ff(5);	/* actual channel doesn't matter */
 
-			set_dma_mode(devpriv->dma1, DMA_MODE_READ);
-			set_dma_addr(devpriv->dma1, virt_to_bus(devpriv->dmaptr1));
-			set_dma_count(devpriv->dma1, n << 1);
-
-			enable_dma(devpriv->dma1);
-			restore_flags(flags);
-		}
-		if (devpriv->ntrig > 0) {
-			n = devpriv->ntrig;
-			if (n > devpriv->dma_maxsize)
-				n = devpriv->dma_maxsize;
-			devpriv->ntrig -= n;
-			devpriv->dma2_size = n;
-			devpriv->current_dma_chan = 0;
-/* DMA junk */
-			{
-				long flags;
-				save_flags(flags);
-				cli();
-				clear_dma_ff(5);	/* actual channel doesn't matter */
-
-				set_dma_mode(devpriv->dma2, DMA_MODE_READ);
-				set_dma_addr(devpriv->dma2, virt_to_bus(devpriv->dmaptr2));
-				set_dma_count(devpriv->dma2, n << 1);
-
-				enable_dma(devpriv->dma2);
-				restore_flags(flags);
-			}
-		}
 		devpriv->adcsr = DT2821_ADCLK | DT2821_IADDONE;
 		update_adcsr(0);
 
@@ -614,11 +689,11 @@ static int dt282x_ao(comedi_device * dev, comedi_subdevice * s, comedi_trig * it
 		/* select channel */
 		devpriv->dacsr |= DT2821_YSEL;
 		if (devpriv->da0_2scomp)
-			data ^= boardtype.dasignbit;
+			data ^= (1<<(boardtype.dabits-1));
 	} else {
 		devpriv->dacsr &= ~DT2821_YSEL;
 		if (devpriv->da1_2scomp)
-			data ^= boardtype.dasignbit;
+			data ^= (1<<(boardtype.dabits-1));
 	}
 
 	update_dacsr(0);
@@ -632,102 +707,102 @@ static int dt282x_ao(comedi_device * dev, comedi_subdevice * s, comedi_trig * it
 
 static int dt282x_dio(comedi_device * dev, comedi_subdevice * s, comedi_trig * it)
 {
-#if 0
-	switch (it->io) {
-	case COMEDI_IO_INPUT:
-		it->data = inw(dev->iobase + DT2821_DIODAT);
+	if(it->flags&TRIG_CONFIG){
+		int mask,i;
 
-		return 0;
-	case COMEDI_IO_OUTPUT:
-		outw(it->data, dev->iobase + DT2821_DIODAT);
-		return 0;
-	default:
-		return -EINVAL;
-	}
-#else
-	return -EINVAL;
-#endif
-}
+		for(i=0;i<it->n_chan;i++){
+			mask=(CR_CHAN(it->chanlist[i])<8)?0x00ff:0xff00;
+			if(it->data[i])s->io_bits|=mask;
+			else s->io_bits&=~mask;
+		}
+		if(s->io_bits&0x00ff)devpriv->dacsr|=DT2821_LBOE;
+		else devpriv->dacsr&=~DT2821_LBOE;
+		if(s->io_bits&0xff00)devpriv->dacsr|=DT2821_HBOE;
+		else devpriv->dacsr&=~DT2821_HBOE;
 
-#if 0
-static int dt282x_dsp(comedi_device * dev, comedi_param * it)
-{
-	switch (it->pnum) {
-	case COMEDI_IOBITS:
-		switch (it->pval & 0xff) {
-		case 0xff & COMEDI_IOBITS_INPUT:
-			devpriv->dacsr &= ~DT2821_LBOE;
-			break;
-		case 0xff & COMEDI_IOBITS_OUTPUT:
-			devpriv->dacsr |= DT2821_LBOE;
-			break;
-		default:
-			return -EINVAL;
-		}
-		switch (it->pval & 0xff00) {
-		case 0xff00 & COMEDI_IOBITS_INPUT:
-			devpriv->dacsr &= ~DT2821_HBOE;
-			break;
-		case 0xff00 & COMEDI_IOBITS_OUTPUT:
-			devpriv->dacsr |= DT2821_HBOE;
-			break;
-		default:
-			return -EINVAL;
-		}
 		outw(devpriv->dacsr, dev->iobase + DT2821_DACSR);
-		changeparam(&dev->chinfo[it->chan].paramlist, COMEDI_IOBITS, 0,
-			    it->pval);
-		return 0;
-	default:
-		return -EINVAL;
+	}else{
+		unsigned int data;
+
+		if(it->flags&TRIG_WRITE){
+			do_pack(&s->state,it);
+			outw(s->state, dev->iobase + DT2821_DIODAT);
+		}else{
+			data = inw(dev->iobase + DT2821_DIODAT);
+			di_unpack(data,it);
+		}
+	}
+
+	return it->n_chan;
+}
+
+
+static int ai_range_table[]={ RANGE_dt282x_ai_lo_bipolar,
+	RANGE_dt282x_ai_lo_unipolar, RANGE_dt282x_ai_5_bipolar,
+	RANGE_dt282x_ai_5_unipolar };
+static int ai_range_pgl_table[]={ RANGE_dt282x_ai_hi_bipolar,
+	RANGE_dt282x_ai_hi_unipolar };
+static inline int opt_ai_range_lkup(int ispgl,int x)
+{
+	if(ispgl){
+		if(x<0 || x>=2)return RANGE_unknown;
+		return ai_range_pgl_table[x];
+	}else{
+		if(x<0 || x>=4)return RANGE_unknown;
+		return ai_range_table[x];
 	}
 }
-#endif
+static int ao_range_table[]={ RANGE_bipolar10, RANGE_unipolar10, RANGE_bipolar5,
+	RANGE_unipolar5, RANGE_bipolar2_5 };
+static inline int opt_ao_range_lkup(int x)
+	{ if(x<0)x=0; if(x>=5)x=0; return ao_range_table[x]; }
+
+enum{	opt_iobase=0, opt_irq, opt_dma1, opt_dma2,	/* i/o base, irq, dma channels */
+	opt_diff,					/* differential */
+	opt_ai_twos, opt_ao0_twos, opt_ao1_twos, 	/* twos comp */
+	opt_ai_range, opt_ao0_range, opt_ao1_range,	/* range */
+};
 
 /*
    options:
-   i/o base
-   irq
-   1=differential, 0=single ended
-   ai 0=unipolar, 1=bipolar
-   ao0 0=unipolar, 1=bipolar
-   ao1 0=unipolar, 1=bipolar
-   dma1
-   dma2
+   0	i/o base
+   1	irq
+   2	dma1
+   3	dma2
+   4	0=single ended, 1=differential
+   5	ai 0=straight binary, 1=2's comp
+   6	ao0 0=straight binary, 1=2's comp
+   7	ao1 0=straight binary, 1=2's comp
+   8	ai 0=10 V, 1=0-10 V, 2=5 V, 3=0-5 V
+   9	ao0 0=10 V, 1=0-10 V, 2=5 V, 3=0-5 V, 4=2.5 V
+   10	ao1 0=10 V, 1=0-10 V, 2=5 V, 3=0-5 V, 4=2.5 V
  */
-
 int dt282x_init(comedi_device * dev, comedi_devconfig * it)
 {
 	int i, irqs, irq;
 	long flags;
 	int board;
-	int board_number;
 	int ret;
 	comedi_subdevice *s;
 
-	board = -1;
-	for (board_number = 0; board_number < n_boardnames; board_number++) {
-		if (!strcmp(boardnames[board_number].name, it->board_name)) {
-			if (it->options[2])
-				board = boardnames[board_number].boardcode_di;
-			else
-				board = boardnames[board_number].boardcode_se;
-			dev->board_name = boardnames[board_number].name;
+	for (board = 0; board < n_boardtypes; board++) {
+		if (!strcmp(boardtypes[board].name, it->board_name)) {
 			break;
 		}
 	}
-	if (board < 0)
+	if (board == n_boardtypes)
 		return 0;
+	dev->board_name = boardtypes[board].name;
 	dev->driver_name = "dt282x";
 
-	if (it->options[0])
-		dev->iobase = it->options[0];
+	if (it->options[opt_iobase])
+		dev->iobase = it->options[opt_iobase];
 	else
 		dev->iobase = 0x240;
 
-	printk("comedi%d: dt282x: 0x%04x\n", dev->minor, dev->iobase);
+	printk("comedi%d: dt282x: 0x%04x", dev->minor, dev->iobase);
 	if (check_region(dev->iobase, DT2821_SIZE) < 0) {
-		comedi_error(dev, "I/O port conflict");
+		printk(" I/O port conflict\n");
 		return -EIO;
 	}
 	request_region(dev->iobase, DT2821_SIZE, "dt282x");
@@ -736,7 +811,7 @@ int dt282x_init(comedi_device * dev, comedi_devconfig * it)
 	outw(DT2821_BDINIT, dev->iobase + DT2821_SUPCSR);
 	i = inw(dev->iobase + DT2821_ADCSR);
 #ifdef DEBUG
-	printk("fingerprint=%x,%x,%x,%x,%x",
+	printk(" fingerprint=%x,%x,%x,%x,%x",
 	       inw(dev->iobase + DT2821_ADCSR),
 	       inw(dev->iobase + DT2821_CHANCSR),
 	       inw(dev->iobase + DT2821_DACSR),
@@ -755,13 +830,13 @@ int dt282x_init(comedi_device * dev, comedi_devconfig * it)
 		  != DT2821_SUPCSR_VAL) ||
 		 ((inw(dev->iobase + DT2821_TMRCTR) & DT2821_TMRCTR_MASK)
 		  != DT2821_TMRCTR_VAL)) {
-		comedi_error(dev, "board not found\n");
+		printk(" board not found");
 		ret= -EIO;
 		goto cleanup;
 	}
 	/* should do board test */
 
-	irq = it->options[1];
+	irq = it->options[opt_irq];
 	if (irq < 0) {
 		save_flags(flags);
 		sti();
@@ -774,23 +849,29 @@ int dt282x_init(comedi_device * dev, comedi_devconfig * it)
 		irq = probe_irq_off(irqs);
 		restore_flags(flags);
 		if (0 /* error */ ) {
-			comedi_error(dev, "error probing irq (bad)");
+			printk(" error probing irq (bad)");
 		}
 	}
 	dev->irq = 0;
 	if (irq > 0) {
-		printk("( irq = %d )\n", irq);
+		printk(" ( irq = %d )", irq);
 		request_irq(irq, dt282x_interrupt, SA_INTERRUPT, "dt282x", dev);
 		dev->irq = irq;
 	} else if (irq == 0) {
-		printk("(no irq)\n");
+		printk(" (no irq)");
 	} else {
-		printk("(probe returned multiple irqs--bad)\n");
+		printk(" (probe returned multiple irqs--bad)");
 	}
 
 	if((ret=alloc_private(dev,sizeof(dt282x_private)))<0)
 		goto cleanup;
 
+#if DT_DMA
+	ret=dt282x_grab_dma(dev,it->options[opt_dma1],it->options[opt_dma2]);
+	if(ret<0)
+		goto cleanup;
+#endif
+
 	dev->n_subdevices = 3;
 	if((ret=alloc_subdevices(dev))<0)
 		goto cleanup;
@@ -801,37 +882,35 @@ int dt282x_init(comedi_device * dev, comedi_devconfig * it)
 
 	/* ai subdevice */
 	s->type=COMEDI_SUBD_AI;
-	s->subdev_flags=SDF_READABLE;
-	s->n_chan=boardtype.adchans;
+	s->subdev_flags=SDF_READABLE|((it->options[opt_diff])?SDF_DIFF:SDF_COMMON);
+	s->n_chan=(it->options[opt_diff])?boardtype.adchan_di:boardtype.adchan_se;
 	s->trig[0]=dt282x_ai_mode0;
 	s->trig[1]=dt282x_ai_mode1;
 	s->trig[2]=dt282x_ai_mode2;
-	s->maxdata=boardtype.admask;
+	s->maxdata=(1<<boardtype.adbits)-1;
 	s->len_chanlist=16;
-	switch (it->options[3]) {
-	case 0:
-		s->range_type = boardnames[board_number].t0;
-		break;
-	case 1:
-		s->range_type = boardnames[board_number].t1;
-		break;
-	case 2:
-		s->range_type = boardnames[board_number].t2;
-		break;
-	default:
-		s->range_type = 0;
-	}
+	s->range_type = opt_ai_range_lkup(boardtype.ispgl,it->options[opt_ai_range]);
 	s->timer_type=TIMER_dt282x;
+	devpriv->ad_2scomp=it->options[opt_ai_twos];
 
 	s++;
-	/* ao subsystem */
-	s->type=COMEDI_SUBD_AO;
-	s->subdev_flags=SDF_WRITEABLE;
-	s->n_chan=boardtype.dachans;
-	s->trig[0]=dt282x_ao;
-	s->maxdata=boardtype.damask;
-	s->len_chanlist=1;			/* XXX can do 2 */
-	s->range_type = RANGE_bipolar10;	/* XXX wrong */
+	if((s->n_chan=boardtype.dachan)){
+		/* ao subsystem */
+		s->type=COMEDI_SUBD_AO;
+		s->subdev_flags=SDF_WRITEABLE;
+		s->trig[0]=dt282x_ao;
+		s->maxdata=(1<<boardtype.dabits)-1;
+		s->len_chanlist=1;			/* XXX could do 2 */
+		s->range_type_list=devpriv->darangelist;
+		devpriv->darangelist[0]=
+			opt_ao_range_lkup(it->options[opt_ao0_range]);
+		devpriv->darangelist[1]=
+			opt_ao_range_lkup(it->options[opt_ao1_range]);
+		devpriv->da0_2scomp=it->options[opt_ao0_twos];
+		devpriv->da1_2scomp=it->options[opt_ao1_twos];
+	}else{
+		s->type=COMEDI_SUBD_UNUSED;
+	}
 
 	s++;
 	/* dio subsystem */
@@ -842,18 +921,13 @@ int dt282x_init(comedi_device * dev, comedi_devconfig * it)
 	s->maxdata=1;
 	s->range_type = RANGE_digital;
 
-#if DT_DMA
-	/*
- 	*    DMA is corrupting memory.  disabled.
- 	*/
-	if(ret=dt282x_grab_dma(dev,it->options[6],it->options[7]))
-		goto cleanup;
-#endif
-
 	dev->rem = dt282x_rem;
 
+	printk("\n");
+
 	return 1;
 cleanup:
+	printk("\n");
 	free_resources(dev);
 	return ret;
 }
@@ -875,10 +949,7 @@ static void free_resources(comedi_device *dev)
 			free_page((unsigned long) devpriv->dmaptr1);
 		if (devpriv->dmaptr2)
 			free_page((unsigned long) devpriv->dmaptr2);
-		kfree(dev->private);
 	}
-	if(dev->subdevices)
-		kfree(dev->subdevices);
 }
 
 static int dt282x_rem(comedi_device * dev)
@@ -891,12 +962,6 @@ static int dt282x_rem(comedi_device * dev)
 }
 
 
-/*
- *    Start of the DMA allocation atrocity.
- *      This section screams for a rewrite.
- *	rewritten.  looks better now.
- */
-
 #if DT_DMA
 static int dt282x_grab_dma(comedi_device *dev,int dma1,int dma2)
 {
@@ -904,6 +969,11 @@ static int dt282x_grab_dma(comedi_device *dev,int dma1,int dma2)
 
 	devpriv->usedma=0;
 
+	if(!dma1 && !dma2){
+		printk(" (no dma)");
+		return 0;
+	}
+
 	if(dma1==dma2 || dma1<5 || dma2<5 || dma1>7 || dma2>7)
 		return -EINVAL;
 
@@ -911,12 +981,12 @@ static int dt282x_grab_dma(comedi_device *dev,int dma1,int dma2)
 		int i;i=dma1;dma1=dma2;dma2=i;
 	}
 
-	ret = request_dma(dma1, "dt282x");
+	ret = request_dma(dma1, "dt282x A");
 	if (ret)
 		return -EBUSY;
 	devpriv->dma1=dma1;
 
-	ret = request_dma(dma2, "dt282x");
+	ret = request_dma(dma2, "dt282x B");
 	if (ret)
 		return -EBUSY;
 	devpriv->dma2=dma2;
@@ -925,19 +995,11 @@ static int dt282x_grab_dma(comedi_device *dev,int dma1,int dma2)
 	devpriv->dmaptr1 = (void *) get_free_page(GFP_KERNEL | GFP_DMA);
 	devpriv->dmaptr2 = (void *) get_free_page(GFP_KERNEL | GFP_DMA);
 	if (!devpriv->dmaptr1 || !devpriv->dmaptr2) {
-		printk("can't get DMA memory ");
+		printk(" can't get DMA memory");
 		return -ENOMEM;
 	}
 
-	{
-	/* for debugging */
-	int i;
-
-	for (i = 0; i < devpriv->dma_maxsize; i++) {
-		devpriv->dmaptr1[i] = 0xdead;
-		devpriv->dmaptr2[i] = 0xdead;
-	}
-	}
+	printk(" (dma=%d,%d)",dma1,dma2);
 
 	devpriv->usedma=1;
 
