@@ -1,6 +1,6 @@
 /*
-    module/module.c
-    comedi kernel module
+    module/exp_ioctl.c
+    exported comedi functions
 
     COMEDI - Linux Control and Measurement Device Interface
     Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>
@@ -55,12 +55,6 @@ int comedi_subdinfo_ioctl(unsigned int minor,comedi_subdinfo *arg,void *file);
 int comedi_chaninfo_ioctl(unsigned int minor,comedi_chaninfo *arg);
 #endif
 
-int comedi_trig_ioctl(unsigned int minor,unsigned int subdev,comedi_trig *it);
-int comedi_lock_ioctl(unsigned int minor,unsigned int subdev);
-int comedi_unlock_ioctl(unsigned int minor,unsigned int subdev);
-int comedi_cancel_ioctl(unsigned int minor,unsigned int subdev);
-int comedi_register_callback(unsigned int minor,unsigned int subdev,
-		unsigned int type,int (*cb)(void *),void *arg);
 
 
 static inline int minor_to_dev(unsigned int minor,comedi_device **dev)
@@ -218,7 +212,7 @@ static int do_chaninfo_ioctl(comedi_device *dev,comedi_chaninfo *arg)
 			
 	if(it.rangelist){
 		if(s->subdev_flags & SDF_FLAGS)
-			ret=copy_to_user(it.rangelist,s->range_list,s->n_chan*sizeof(unsigned int));
+			ret=copy_to_user(it.rangelist,s->range_type_list,s->n_chan*sizeof(unsigned int));
 		else
 			ret=clear_user(it.rangelist,s->n_chan*sizeof(unsigned int));
 		if(ret)return -EFAULT;
@@ -250,7 +244,7 @@ int comedi_trig_ioctl(unsigned int minor,unsigned int subdev,comedi_trig *it)
 {
 	comedi_device *dev;
 	comedi_subdevice *s;
-	int ret=0,i;
+	int ret=0;
 
 	if((ret=minor_to_dev(minor,&dev))<0)
 		return ret;
@@ -288,41 +282,39 @@ int comedi_trig_ioctl(unsigned int minor,unsigned int subdev,comedi_trig *it)
 		goto cleanup;
 	
 	s->buf_user_ptr=0;
+	s->buf_user_count=0;
 	s->buf_int_ptr=0;
-	
+	s->buf_int_count=0;
+
 	if(it->data==NULL){
 		ret=-EINVAL;
 		goto cleanup;
 	}
 
+	if(!it->data_len){
+#if 0
+		ret=-EINVAL;
+		goto cleanup;
+#else
+		it->data_len=it->n_chan*it->n;
+		rt_printk("comedi: warning: trig->data_len not set\n");
+#endif
+	}
+
 	if(it->mode>=5 || s->trig[it->mode]==NULL){
 		ret=-EINVAL;
 		goto cleanup;
 	}
 
-	/* mark as an RT operation */
-	it->flags |= TRIG_RT;
-
-	/* set EOS flag */
-	if(s->end_of_scan){
-		it->flags |= TRIG_WAKE_EOS;
-		printk("setting eos flag\n");
-	}
-
 	s->cur_trig=*it;
 	
 	ret=s->trig[it->mode](dev,s,it);
 	
-	if(ret==0)return 0;  /* driver now owns kmalloc'd areas */
+	if(ret==0)return 0;
 	if(ret<0)goto cleanup;
 
-	i=ret*sizeof(sampl_t);
-	if(i>it->n*it->n_chan){
-#if 0
-		/* can we call printk from RT? */
-		printk("comedi: (bug) trig returned too many samples\n");
-#endif
-		i=it->n*it->n_chan;
+	if(ret>it->n*it->n_chan){
+		rt_printk("comedi: (bug) trig returned too many samples\n");
 	}
 
 cleanup:
@@ -331,6 +323,26 @@ cleanup:
 	return ret;
 }
 
+/*
+   This function does the same as above, but without any sanity
+   checks.  Any insanity in code calling this function must be
+   assumed by the writer.
+ */
+int __comedi_trig_ioctl(unsigned int minor,unsigned int subdev,comedi_trig *it)
+{
+	comedi_device *dev;
+	comedi_subdevice *s;
+	int ret=0;
+
+	dev=comedi_devices+minor;
+	s=dev->subdevices+subdev;
+
+	s->cur_trig=*it;
+	
+	ret=s->trig[it->mode](dev,s,it);
+	
+	return ret;
+}
 
 /*
 	COMEDI_LOCK
@@ -429,8 +441,9 @@ int comedi_unlock_ioctl(unsigned int minor,unsigned int subdev)
 	if(s->lock==&rtcomedi_lock_semaphore){
 		s->lock=NULL;
 
-		s->end_of_scan=NULL;
-		s->end_of_acquisition=NULL;
+		s->cb_mask=0;
+		s->cb_func=NULL;
+		s->cb_arg=NULL;
 	}
 
 	return 0;
@@ -492,7 +505,8 @@ int comedi_cancel_ioctl(unsigned int minor,unsigned int subdev)
    XXX - this needs additional work.  Specifically, being SDF_RT is _not_ a
    sufficient condition for being able to do callbacks.
  */
-int comedi_register_callback(unsigned int minor,unsigned int subdev,unsigned int type,int (*cb)(void *),void *arg)
+int comedi_register_callback(unsigned int minor,unsigned int subdev,
+		unsigned int mask,int (*cb)(unsigned int,void *),void *arg)
 {
 	comedi_device *dev;
 	comedi_subdevice *s;
@@ -520,18 +534,17 @@ int comedi_register_callback(unsigned int minor,unsigned int subdev,unsigned int
 	if(s->busy)
 		return -EBUSY;
 
-	switch(type){
-	case COMEDI_CB_EOS:
-		s->end_of_scan=cb;
-		s->eos_arg=arg;
-		return 0;
-	case COMEDI_CB_EOA:
-		s->end_of_acquisition=cb;
-		s->eoa_arg=arg;
-		return 0;
-	default:
-		return -EINVAL;
+	if(!mask){
+		s->cb_mask=0;
+		s->cb_func=NULL;
+		s->cb_arg=NULL;
+	}else{
+		s->cb_mask=mask;
+		s->cb_func=cb;
+		s->cb_arg=arg;
 	}
+
+	return 0;
 }
 
 
@@ -555,6 +568,7 @@ int comedi_register_callback(unsigned int minor,unsigned int subdev,unsigned int
 struct symbol_table comedi_syms = {
 #include <linux/symtab_begin.h>
 	X(comedi_trig_ioctl),
+	X(__comedi_trig_ioctl),
 	X(comedi_lock_ioctl),
 	X(comedi_unlock_ioctl),
 	X(comedi_cancel_ioctl),
@@ -569,6 +583,7 @@ struct symbol_table comedi_syms = {
 #ifdef LINUX_V22
 
 EXPORT_SYMBOL(comedi_trig_ioctl);
+EXPORT_SYMBOL(__comedi_trig_ioctl);
 EXPORT_SYMBOL(comedi_lock_ioctl);
 EXPORT_SYMBOL(comedi_unlock_ioctl);
 EXPORT_SYMBOL(comedi_cancel_ioctl);
