@@ -40,6 +40,9 @@
 #ifdef LINUX_V22
 #include <asm/uaccess.h>
 #endif
+#ifdef HAVE_RTL_POSIXIO
+#include <rtlinux/rtl_posixio.h>
+#endif
 
 comedi_device *comedi_devices;
 
@@ -60,6 +63,7 @@ static int comedi_ioctl(struct inode * inode,struct file * file,unsigned int cmd
 {
 	kdev_t minor=MINOR(inode->i_rdev);
 	comedi_device *dev=comedi_devices+minor;
+
 	
 	switch(cmd)
 	{
@@ -129,6 +133,8 @@ static int do_devconfig_ioctl(comedi_device *dev,comedi_devconfig *arg,kdev_t mi
 		if(ret<0)
 			return ret;
 		if(ret){
+			init_waitqueue_head(&dev->wait);
+
 			/* do a little post-config cleanup */
 			postconfig(dev);
 			return 0;
@@ -161,6 +167,13 @@ static int do_devinfo_ioctl(comedi_device *dev,comedi_devinfo *arg)
 	/* fill devinfo structure */
 	devinfo.version_code=COMEDI_VERSION_CODE;
 	devinfo.n_subdevs=dev->n_subdevices;
+#if 1
+	if(!dev->driver_name || !dev->board_name){
+		printk("BUG: dev->driver_name=<%p> dev->board_name=<%p>\n",
+			dev->driver_name,dev->board_name);
+		return -EFAULT;
+	}
+#endif
 	memcpy(devinfo.driver_name,dev->driver_name,COMEDI_NAMELEN);
 	memcpy(devinfo.board_name,dev->board_name,COMEDI_NAMELEN);
 	memcpy(devinfo.options,dev->options,COMEDI_NDEVCONFOPTS*sizeof(int));
@@ -198,6 +211,8 @@ static int do_subdinfo_ioctl(comedi_device *dev,comedi_subdinfo *arg,void *file)
 	if(!tmp)
 		return -ENOMEM;
 	
+	memset(tmp,0,sizeof(comedi_subdinfo)*dev->n_subdevices);
+
 	/* fill subdinfo structs */
 	for(i=0;i<dev->n_subdevices;i++){
 		s=dev->subdevices+i;
@@ -210,6 +225,7 @@ static int do_subdinfo_ioctl(comedi_device *dev,comedi_subdinfo *arg,void *file)
 		us->len_chanlist	= s->len_chanlist;
 		us->maxdata		= s->maxdata;
 		us->range_type		= s->range_type;
+		us->flags		= s->flags;
 		
 		if(s->busy)
 			us->subd_flags |= SDF_BUSY;
@@ -219,7 +235,7 @@ static int do_subdinfo_ioctl(comedi_device *dev,comedi_subdinfo *arg,void *file)
 			us->subd_flags |= SDF_LOCKED;
 		if(s->lock == file)
 			us->subd_flags |= SDF_LOCK_OWNER;
-		if(s->maxdata_list)
+		if(!s->maxdata && s->maxdata_list)
 			us->subd_flags |= SDF_MAXDATA;
 		if(s->flaglist)
 			us->subd_flags |= SDF_FLAGS;
@@ -263,7 +279,6 @@ static int do_chaninfo_ioctl(comedi_device *dev,comedi_chaninfo *arg)
 {
 	comedi_subdevice *s;
 	comedi_chaninfo it;
-	int ret;
 	
 	if(copy_from_user(&it,arg,sizeof(comedi_chaninfo)))
 		return -EFAULT;
@@ -272,20 +287,23 @@ static int do_chaninfo_ioctl(comedi_device *dev,comedi_chaninfo *arg)
 		return -EINVAL;
 	s=dev->subdevices+it.subdev;
 	
+	if(it.maxdata_list){
+		if(s->maxdata || !s->maxdata_list)
+			return -EINVAL;
+		if(copy_to_user(it.maxdata_list,s->maxdata_list,s->n_chan*sizeof(lsampl_t)))
+			return -EFAULT;
+	}
+
 	if(it.flaglist){
-		if(s->subdev_flags & SDF_FLAGS)
-			ret=copy_to_user(it.flaglist,s->flaglist,s->n_chan*sizeof(unsigned int));
-		else
-			ret=clear_user(it.flaglist,s->n_chan*sizeof(unsigned int));
-		if(ret)return -EFAULT;
+		if(!s->flaglist)return -EINVAL;
+		if(copy_to_user(it.flaglist,s->flaglist,s->n_chan*sizeof(unsigned int)))
+			return -EFAULT;
 	}
 			
 	if(it.rangelist){
-		if(s->subdev_flags & SDF_FLAGS)
-			ret=copy_to_user(it.rangelist,s->range_list,s->n_chan*sizeof(unsigned int));
-		else
-			ret=clear_user(it.rangelist,s->n_chan*sizeof(unsigned int));
-		if(ret)return -EFAULT;
+		if(!s->range_type_list)return -EINVAL;
+		if(copy_to_user(it.rangelist,s->range_type_list,s->n_chan*sizeof(unsigned int)))
+			return -EFAULT;
 	}
 	
 	return 0;
@@ -314,9 +332,13 @@ static int do_trig_ioctl(comedi_device *dev,void *arg,void *file)
 	comedi_trig user_trig;
 	comedi_subdevice *s;
 	int ret=0,i,bufsz;
+	int reading;
 	
-	if(copy_from_user(&user_trig,arg,sizeof(comedi_trig)))
+DPRINTK("entering do_trig_ioctl()\n");
+	if(copy_from_user(&user_trig,arg,sizeof(comedi_trig))){
+		DPRINTK("bad trig address\n");
 		return -EFAULT;
+	}
 	
 #if 0
 	/* this appears to be the only way to check if we are allowed
@@ -325,20 +347,28 @@ static int do_trig_ioctl(comedi_device *dev,void *arg,void *file)
 		return -EFAULT;
 #endif
 	
-	if(user_trig.subdev>=dev->n_subdevices)
+	if(user_trig.subdev>=dev->n_subdevices){
+		DPRINTK("%d no such subdevice\n",user_trig.subdev);
 		return -ENODEV;
+	}
 
 	s=dev->subdevices+user_trig.subdev;
-	if(s->type==COMEDI_SUBD_UNUSED)
+	if(s->type==COMEDI_SUBD_UNUSED){
+		DPRINTK("%d not used device\n",user_trig.subdev);
 		return -EIO;
+	}
 	
 	/* are we locked? (ioctl lock) */
-	if(s->lock && s->lock!=file)
+	if(s->lock && s->lock!=file){
+		DPRINTK("device locked\n");
 		return -EACCES;
+	}
 
 	/* are we busy? */
-	if(s->busy)
+	if(s->busy){
+		DPRINTK("device busy\n");
 		return -EBUSY;
+	}
 	s->busy=file;
 
 	/* make sure channel/gain list isn't too long */
@@ -356,46 +386,57 @@ static int do_trig_ioctl(comedi_device *dev,void *arg,void *file)
 	/* load channel/gain list */
 	s->cur_trig.chanlist=kmalloc(s->cur_trig.n_chan*sizeof(int),GFP_KERNEL);
 	if(!s->cur_trig.chanlist){
+		DPRINTK("allocation failed\n");
 		ret = -ENOMEM;
 		goto cleanup;
 	}
 	
 	if(copy_from_user(s->cur_trig.chanlist,user_trig.chanlist,s->cur_trig.n_chan*sizeof(int))){
+		DPRINTK("fault reading chanlist\n");
 		ret = -EFAULT;
 		goto cleanup;
 	}
 	
 	/* make sure each element in channel/gain list is valid */
-	if((ret=check_chanlist(s,s->cur_trig.n_chan,s->cur_trig.chanlist))<0)
+	if((ret=check_chanlist(s,s->cur_trig.n_chan,s->cur_trig.chanlist))<0){
+		DPRINTK("bad chanlist\n");
 		goto cleanup;
+	}
 	
-	/* allocate buffer */
-	if(s->flags&SDF_LSAMPL){
-		bufsz=s->cur_trig.n*s->cur_trig.n_chan*sizeof(lsampl_t);
-	}else{
-		bufsz=s->cur_trig.n*s->cur_trig.n_chan*sizeof(sampl_t);
-	}
-	if(bufsz>s->maxbufsz){
-		DPRINTK("buffer size too big %d > %d\n",bufsz,s->maxbufsz);
-		/* XXX we *should* be able to handle this */
-		ret=-EINVAL;
-		goto cleanup;
-	}
-	if(s->buf){
-		s->cur_trig.data=s->buf;
-	}else{
+	if(!s->prealloc_bufsz){
+		/* allocate temporary buffer */
+
+		if(s->subdev_flags&SDF_LSAMPL){
+			bufsz=s->cur_trig.n*s->cur_trig.n_chan*sizeof(lsampl_t);
+		}else{
+			bufsz=s->cur_trig.n*s->cur_trig.n_chan*sizeof(sampl_t);
+		}
+
 		if(!(s->cur_trig.data=kmalloc(bufsz,GFP_KERNEL))){
+			DPRINTK("failed to allocate buffer\n");
 			ret=-ENOMEM;
 			goto cleanup;
 		}
+	}else{
+		bufsz=s->prealloc_bufsz;
+		if(!s->prealloc_buf){
+			printk("comedi: bug: s->prealloc_buf=NULL\n");
+		}
+		s->cur_trig.data=s->prealloc_buf;
 	}
+	s->cur_trig.data_len=bufsz;
 
+#if 0
 /* debugging */
 memset(s->cur_trig.data,0xef,bufsz);
+#endif
 
 	s->buf_int_ptr=0;
+	s->buf_int_count=0;
+if(s->subdev_flags & SDF_READABLE){
 	s->buf_user_ptr=0;
-	s->buf_len=s->cur_trig.n*s->cur_trig.n_chan;
+	s->buf_user_count=0;
+}
 	
 #if 0
 	if(user_trig.data==NULL){
@@ -404,7 +445,23 @@ memset(s->cur_trig.data,0xef,bufsz);
 		goto cleanup;
 	}
 #endif
+
 	if(s->subdev_flags & SDF_WRITEABLE){
+		if(s->subdev_flags & SDF_READABLE){
+			/* bidirectional, so we defer to trig structure */
+			if(user_trig.flags&TRIG_WRITE){
+				reading=0;
+			}else{
+				reading=1;
+			}
+		}else{
+			reading=0;
+		}
+	}else{
+		/* subdev is read-only */
+		reading=1;
+	}
+	if(!reading && user_trig.data){
 		if(copy_from_user(s->cur_trig.data,user_trig.data,bufsz)){
 			ret=-EFAULT;
 			goto cleanup;
@@ -428,7 +485,7 @@ memset(s->cur_trig.data,0xef,bufsz);
 
 	if(ret<0)goto cleanup;
 
-	if(s->flags&SDF_LSAMPL){
+	if(s->subdev_flags&SDF_LSAMPL){
 		i=ret*sizeof(lsampl_t);
 	}else{
 		i=ret*sizeof(sampl_t);
@@ -437,7 +494,7 @@ memset(s->cur_trig.data,0xef,bufsz);
 		printk("comedi: (bug) trig returned too many samples\n");
 		i=bufsz;
 	}
-	if(s->subdev_flags & SDF_READABLE){
+	if(reading){
 		if(copy_to_user(user_trig.data,s->cur_trig.data,i)){
 			ret=-EFAULT;
 			goto cleanup;
@@ -503,6 +560,14 @@ static int do_lock_ioctl(comedi_device *dev,unsigned int arg,void * file)
 	
 	rtcomedi_lock_semaphore=0;
 
+	if(ret<0)
+		return ret;
+
+#if 0
+	if(s->lock_f)
+		ret=s->lock_f(dev,s);
+#endif
+
 	return ret;
 }
 
@@ -537,8 +602,14 @@ static int do_unlock_ioctl(comedi_device *dev,unsigned int arg,void * file)
 	if(s->lock && s->lock!=file)
 		return -EACCES;
 	
-	if(s->lock==file)
+	if(s->lock==file){
+#if 0
+		if(s->unlock)
+			s->unlock(dev,s);
+#endif
+
 		s->lock=NULL;
+	}
 
 	return 0;
 }
@@ -601,10 +672,11 @@ static int do_cancel(comedi_device *dev,comedi_subdevice *s)
 
    unmapping:
    	vm_ops->unmap()
+	- this needs to call comedi_cancel, or whatever.
  */
 static int comedi_mmap_v22(struct file * file, struct vm_area_struct *vma)
 {
-	kdev_t minor=minor_of_file(file);
+	kdev_t minor=MINOR(RDEV_OF_FILE(file));
 	comedi_device *dev=comedi_devices+minor;
 	comedi_subdevice *s;
 	int size;
@@ -613,52 +685,138 @@ static int comedi_mmap_v22(struct file * file, struct vm_area_struct *vma)
 		return -EIO;
 	s=dev->subdevices+vma->vm_offset;
 
-	if(!(s->subdev_flags&SDF_MMAP))
-		return -EIO;
-
 	if((vma->vm_flags & VM_WRITE) && !(s->subdev_flags & SDF_WRITEABLE))
 		return -EINVAL;
 
 	if((vma->vm_flags & VM_READ) && !(s->subdev_flags & SDF_READABLE))
 		return -EINVAL;
 
-	/* check if aleady mmaped */
-
 	size = vma->vm_end - vma->vm_start;
+	if(size>(1<<16))
+		return -EINVAL;
 
-	/* do some size checking */
-
-	if(remap_page_range(vma->vm_start, virt_to_phys(s->buf),
+	if(remap_page_range(vma->vm_start, virt_to_phys(s->prealloc_buf),
 		size,vma->vm_page_prot))
 		return -EAGAIN;
 	
 	vma->vm_file=file;
+#if 0
 	file->f_count++;
+#else
+	file_atomic_inc(&file->f_count);
+#endif
 
 	/* mark subdev as mapped */
 	
 	/* call subdev about mmap, if necessary */
-
-	memset(s->buf,0,size);
+printk("mmap done\n");
 
 	return 0;
 }
+
+#if 0
+/*
+   I can't find a driver that notices when it gets unmapped.
+ */
+static void *comedi_unmap(struct vm_area_struct *area,unsigned long x,size_t y)
+{
+	printk("comedi unmap\n");
+
+	return NULL;
+}
+#endif
+
 #endif
 
 
 static ssize_t comedi_write_v22(struct file *file,const char *buf,size_t nbytes,loff_t *offset)
 {
-#if 0
-	int retval;
-	kdev_t minor=minor_of_file(file);
-	comedi_device *dev=comedi_devices+minor;
-#endif
+	comedi_device *dev;
+	comedi_subdevice *s;
+	int n,m,count=0,retval=0;
+	DECLARE_WAITQUEUE(wait,current);
+	int sample_size;
+	void *buf_ptr;
+	unsigned int buf_len;
+
+	dev=comedi_devices+MINOR(RDEV_OF_FILE(file));
+	s=dev->subdevices+file->f_pos;
+
+	if(s->subdev_flags&SDF_LSAMPL){
+		sample_size=sizeof(lsampl_t);
+	}else{
+		sample_size=sizeof(sampl_t);
+	}
+	if(nbytes%sample_size)
+		nbytes-=nbytes%sample_size;
 
 	if(!nbytes)return 0;
 
-	/* read data */
-	
-	return -EIO;
+	if(!(s->subdev_flags&SDF_WRITEABLE))
+		return -EIO;
+
+	if(!s->busy){
+		buf_ptr=s->prealloc_buf;
+		buf_len=s->prealloc_bufsz;
+	}else{
+		if(s->busy != file)
+			return -EACCES;
+
+		buf_ptr=s->cur_trig.data;
+		buf_len=s->cur_trig.data_len;
+	}
+
+	if(!buf_ptr)
+		return -EIO;
+
+	add_wait_queue(&dev->wait,&wait);
+	while(nbytes>0 && !retval){
+		current->state=TASK_INTERRUPTIBLE;
+
+		n=nbytes;
+
+		m=buf_len-(s->buf_user_count-s->buf_int_count);
+		if(s->buf_user_ptr+m > buf_len){
+			m=buf_len - s->buf_user_ptr;
+		}
+		if(m<n)n=m;
+
+		if(n==0){
+			if(file->f_flags&O_NONBLOCK){
+				retval=-EAGAIN;
+				break;
+			}
+			if(signal_pending(current)){
+				retval=-ERESTARTSYS;
+				break;
+			}
+			if(!(s->subdev_flags&SDF_RUNNING)){
+				do_become_nonbusy(dev,s);
+				break;
+			}
+			schedule();
+			continue;
+		}
+		m=copy_from_user(buf_ptr+s->buf_user_ptr,buf,n);
+		if(m) retval=-EFAULT;
+		n-=m;
+		
+		count+=n;
+		nbytes-=n;
+		s->buf_user_ptr+=n;
+		s->buf_user_count+=n;
+
+		if(s->buf_user_ptr>=buf_len ){
+			s->buf_user_ptr=0;
+		}
+
+		buf+=n;
+		break;	/* makes device work like a pipe */
+	}
+	current->state=TASK_RUNNING;
+	remove_wait_queue(&dev->wait,&wait);
+
+	return (count ? count : retval);
 }
 
 
@@ -667,13 +825,13 @@ static ssize_t comedi_read_v22(struct file * file,char *buf,size_t nbytes,loff_t
 	comedi_device *dev;
 	comedi_subdevice *s;
 	int n,m,count=0,retval=0;
-	struct wait_queue wait={current,NULL};
+	DECLARE_WAITQUEUE(wait,current);
 	int sample_size;
 
-	dev=comedi_devices+minor_of_file(file);
+	dev=comedi_devices+MINOR(RDEV_OF_FILE(file));
 	s=dev->subdevices+file->f_pos;
 
-	if(s->flags&SDF_LSAMPL){
+	if(s->subdev_flags&SDF_LSAMPL){
 		sample_size=sizeof(lsampl_t);
 	}else{
 		sample_size=sizeof(sampl_t);
@@ -698,7 +856,19 @@ static ssize_t comedi_read_v22(struct file * file,char *buf,size_t nbytes,loff_t
 
 		n=nbytes;
 
-		m=(s->buf_int_ptr-s->buf_user_ptr)*sample_size;
+		m=s->buf_int_count-s->buf_user_count;
+		if(m>s->cur_trig.data_len){
+			s->buf_user_count=s->buf_int_count;
+			s->buf_user_ptr=s->buf_int_ptr;
+			retval=-EINVAL; /* OVERRUN */
+			break;
+		}
+		if(s->buf_user_ptr+m > s->cur_trig.data_len){
+			m=s->cur_trig.data_len - s->buf_user_ptr;
+#if 0
+printk("m is %d\n",m);
+#endif
+		}
 		if(m<n)n=m;
 
 		if(n==0){
@@ -717,19 +887,23 @@ static ssize_t comedi_read_v22(struct file * file,char *buf,size_t nbytes,loff_t
 			schedule();
 			continue;
 		}
-		m=copy_to_user(buf,s->cur_trig.data+s->buf_user_ptr,n);
+		m=copy_to_user(buf,((void *)(s->cur_trig.data))+s->buf_user_ptr,n);
 		if(m) retval=-EFAULT;
 		n-=m;
 		
 		count+=n;
 		nbytes-=n;
-		s->buf_user_ptr+=n/sample_size;
+		s->buf_user_ptr+=n;
+		s->buf_user_count+=n;
 
-		if(s->buf_user_ptr>=s->buf_len){
+		if(s->buf_user_ptr>=s->cur_trig.data_len ){
+			s->buf_user_ptr=0;
+#if 0
 			if(s->subdev_flags & SDF_RUNNING)
 				printk("ack.  end of buffer reached!\n");
 			else
 				do_become_nonbusy(dev,s);
+#endif
 		}
 
 		buf+=n;
@@ -758,12 +932,17 @@ static void do_become_nonbusy(comedi_device *dev,comedi_subdevice *s)
 	}
 
 	if(s->cur_trig.data){
-		if(s->cur_trig.data!=s->buf)
+		if(s->cur_trig.data != s->prealloc_buf)
 			kfree(s->cur_trig.data);
 
 		s->cur_trig.data=NULL;
 	}
 
+	s->buf_user_ptr=0;
+	s->buf_int_ptr=0;
+	s->buf_user_count=0;
+	s->buf_int_count=0;
+
 	s->busy=NULL;
 }
 
@@ -777,7 +956,7 @@ static loff_t comedi_lseek_v22(struct file *file,loff_t offset,int origin)
 	comedi_device *dev;
 	loff_t new_offset;
 	
-	dev=comedi_devices+minor_of_file(file);
+	dev=comedi_devices+MINOR(RDEV_OF_FILE(file));
 
 	switch(origin){
 	case SEEK_SET:
@@ -800,7 +979,7 @@ static loff_t comedi_lseek_v22(struct file *file,loff_t offset,int origin)
 
 static int comedi_open(struct inode *inode,struct file *file)
 {
-	kdev_t minor=minor_of_file(file);
+	kdev_t minor=MINOR(inode->i_rdev);
 	comedi_device *dev;
 
 	if(minor>=COMEDI_NDEVICES)return -ENODEV;
@@ -818,7 +997,7 @@ static int comedi_open(struct inode *inode,struct file *file)
 
 static int comedi_close_v22(struct inode *inode,struct file *file)
 {
-	comedi_device *dev=comedi_devices+minor_of_file(file);
+	comedi_device *dev=comedi_devices+MINOR(inode->i_rdev);
 	comedi_subdevice *s;
 	int i;
 
@@ -909,7 +1088,9 @@ int init_module(void)
 	if(!comedi_devices)
 		return -ENOMEM;
 	memset(comedi_devices,0,sizeof(comedi_device)*COMEDI_NDEVICES);
+#if 0
 	init_polling();
+#endif
 
 #ifdef CONFIG_EXPORT
 #ifdef LINUX_V20
@@ -920,62 +1101,112 @@ int init_module(void)
 	/* XXX requires /proc interface */
 	comedi_proc_init();
 	
+#ifdef HAVE_RTL_POSIXIO
+	rtl_register_chrdev(COMEDI_MAJOR,"comedi",&comedi_rtl_fops);
+#endif
+
+	rt_printk_init();
+
 	return 0;
 }
 
+void comedi_dev_free(comedi_device *dev)
+{
+	int i;
+	comedi_subdevice *s;
+
+	for(i=0;i<dev->n_subdevices;i++){
+		s=dev->subdevices+i;
+		if(s->prealloc_buf)
+			vfree(s->prealloc_buf);
+	}
+	if(dev->rem)dev->rem(dev);
+	if(dev->subdevices)kfree(dev->subdevices);
+	if(dev->private)kfree(dev->private);
+}
+
 void cleanup_module(void)
 {
 	int i;
 
 	if(MOD_IN_USE)
 		printk("comedi: module in use -- remove delayed\n");
-	comedi_proc_cleanup();
-	for(i=0;i<COMEDI_NDEVICES;i++){
-		if(comedi_devices[i].rem)
-			comedi_devices[i].rem(comedi_devices+i);
-	}
-	cleanup_polling();
-	kfree(comedi_devices);
 
 	unregister_chrdev(COMEDI_MAJOR,"comedi");
+#ifdef HAVE_RTL_POSIXIO
+	rtl_unregister_chrdev(COMEDI_MAJOR,"comedi");
+#endif
+
+	comedi_proc_cleanup();
+#if 0
+	comedi_polling_cleanup();
+#endif
+	for(i=0;i<COMEDI_NDEVICES;i++){
+		if(comedi_devices[i].rem)comedi_dev_free(comedi_devices+i);
+	}
+	kfree(comedi_devices);
+
+	rt_printk_cleanup();
 }
 
 
 void comedi_error(comedi_device *dev,const char *s)
 {
-	printk("comedi%d: %s: %s\n",dev->minor,dev->driver_name,s);
+	rt_printk("comedi%d: %s: %s\n",dev->minor,dev->driver_name,s);
 }
 
 void comedi_done(comedi_device *dev,comedi_subdevice *s)
 {
+#if 0
 	DPRINTK("comedi_done\n");
+#endif
 
 	if(!(s->cur_trig.flags&TRIG_RT))
 		wake_up_interruptible(&dev->wait);
-	else if(s->end_of_acquisition)
-		s->end_of_acquisition(s->eoa_arg);
+	else if(s->cb_mask&COMEDI_CB_EOA)
+		s->cb_func(COMEDI_CB_EOA,s->cb_arg);
 
 	s->subdev_flags &= ~SDF_RUNNING;
 }
 
+void comedi_bufcheck(comedi_device *dev,comedi_subdevice *s)
+{
+#if 0
+	DPRINTK("comedi_bufcheck\n");
+#endif
+
+#if 0
+	if((!(s->cur_trig.flags&TRIG_RT)) &&
+	   (s->buf_int_count-s->buf_user_count >= 16))
+		wake_up_interruptible(&dev->wait);
+#else
+	if(!(s->cur_trig.flags&TRIG_RT))
+		wake_up_interruptible(&dev->wait);
+#endif
+}
+
 /*
    this function should be called by your interrupt routine
    at end-of-scan events
  */
 void comedi_eos(comedi_device *dev,comedi_subdevice *s)
 {
-	if(!(s->cur_trig.flags&TRIG_RT)){
-		if((s->cur_trig.flags&TRIG_WAKE_EOS))
-			wake_up_interruptible(&dev->wait);
-	}else if(s->end_of_scan){
-		int ret;
+	if(s->cb_mask&COMEDI_CB_EOS){
+		s->cb_func(COMEDI_CB_EOS,s->cb_arg);
+	}
+	if((s->cur_trig.flags&TRIG_WAKE_EOS)){
+		wake_up_interruptible(&dev->wait);
+	}
+}
 
-		ret=s->end_of_scan(s->eos_arg);
-		if(ret){
-			/* this is a little hack */
-			s->buf_int_ptr=0;
-			s->cur_chan=0;
-		}
+/*
+   this function should be called by your interrupt routine
+   at buffer rollover events
+ */
+void comedi_eobuf(comedi_device *dev,comedi_subdevice *s)
+{
+	if(s->cb_mask&COMEDI_CB_EOBUF){
+		s->cb_func(COMEDI_CB_EOBUF,s->cb_arg);
 	}
 }
 
@@ -993,8 +1224,20 @@ static void postconfig(comedi_device *dev)
 		if(s->len_chanlist==0)
 			s->len_chanlist=1;
 
-		if(!s->maxbufsz)
-			s->maxbufsz=16*4096;	/* XXX a little low... */
+		/* XXX */
+		if(s->trig[1] || s->trig[2] || s->trig[3] ||s->trig[4]){
+			s->prealloc_bufsz=1024*128;
+		}else{
+			s->prealloc_bufsz=0;
+		}
+
+		if(s->prealloc_bufsz){
+			/* XXX */
+			s->prealloc_buf=vmalloc(s->prealloc_bufsz*sizeof(sampl_t));
+			if(!s->prealloc_buf){
+				printk("ENOMEM\n");
+			}
+		}
 
 		if(!s->range_type && !s->range_type_list)
 			s->range_type=RANGE_unknown;
@@ -1002,3 +1245,34 @@ static void postconfig(comedi_device *dev)
 
 }
 
+
+int di_unpack(unsigned int bits,comedi_trig *it)
+{
+	int chan;
+	int i;
+
+	for(i=0;i<it->n_chan;i++){
+		chan=CR_CHAN(it->chanlist[i]);
+		it->data[i]=(bits>>chan)&1;
+	}
+
+	return i;
+}
+
+int do_pack(unsigned int *bits,comedi_trig *it)
+{
+	int chan;
+	int mask;
+	int i;
+
+	for(i=0;i<it->n_chan;i++){
+		chan=CR_CHAN(it->chanlist[i]);
+		mask=1<<chan;
+		(*bits) &= ~mask;
+		if(it->data[i])
+			(*bits) |=mask;
+	}
+
+	return i;
+}
+
