@@ -3,7 +3,7 @@
     Hardware driver for NI E series cards
 
     COMEDI - Linux Control and Measurement Device Interface
-    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>
+    Copyright (C) 1997-9 David A. Schleef <ds@stm.lbl.gov>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -61,13 +61,15 @@ static int ni_modebits2[4]={ 0x3f, 0x3f, 0x37, 0x37 };
 static int ni_gainlkup[][16]={
 	{ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 },
 	{ 1, 2, 4, 7, 9, 10, 12, 15, 0,0,0,0,0,0,0,0 },
-	{ 1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 12, 13, 14, 15, 0,0 }
+	{ 1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 12, 13, 14, 15, 0,0 },
+	{ 0, 1, 4, 7, 8, 9, 12, 15, 0, 0, 0, 0, 0, 0, 0, 0 }
 };
 
 static int ni_range_lkup[]={
 	RANGE_ni_E_ai,
 	RANGE_ni_E_ai_limited,
-	RANGE_ni_E_ai_limited14
+	RANGE_ni_E_ai_limited14,
+	RANGE_ni_E_ai_limited_602x
 };
 
 /*
@@ -90,10 +92,10 @@ RANGE_ni_E_ai
         0       0.2
         0       0.1
 RANGE_ni_E_ai_limited
+	-10	10
 	-5	5
-	-2.5	2.5
-	-0.5	0.5
-	-0.05	0.05
+	-1	1
+	-0.1	0.1
 	0	10
 	0	5
 	0	1
@@ -113,6 +115,15 @@ RANGE_ni_E_ai_limited14
         0       0.5
         0       0.2
         0       0.1
+RANGE_ni_E_ai_limited_602x
+        -10     10
+        -5      5
+        -0.5    0.5
+        -0.05   0.05
+        0       20
+        0       10
+        0       1
+        0       0.1
 RANGE_ni_E_ao
         -10     10
         0       10
@@ -137,6 +148,8 @@ static int ni_calib(comedi_device *dev,comedi_subdevice *s,comedi_trig *it);
 static void ni_handle_fifo_half_full(comedi_device *dev);
 static void ni_handle_fifo_dregs(comedi_device *dev);
 
+static int ni_ao_fifo_half_empty(comedi_device *dev,comedi_subdevice *s);
+
 static int ni_8255_callback(int dir,int port,int data,void *arg);
 
 #undef TIMEMARK
@@ -146,21 +159,88 @@ static int ni_8255_callback(int dir,int port,int data,void *arg);
 #include <linux/time.h>
 #endif
 
+static void do_interrupt(comedi_device *dev);
+
+#define RTirqmask_AI 1
+
 #ifdef CONFIG_COMEDI_RT
-static void ni_E_interrupt(void)
+#ifdef RTL_V1
+static void ni_E_rt_interrupt(void);
+#else
+static unsigned int ni_E_rt_interrupt(unsigned int irq,struct pt_regs *regs);
+#endif
+static comedi_device *ni_rt_interrupt_dev;
+
+static int grab_rt_irq(comedi_device *dev,int mask)
+{
+	int ret;
+
+	if(!devpriv->rt_irq){
+		ret=rtl_request_global_irq(dev->irq,ni_E_rt_interrupt);
+		if(ret<0){
+			rt_printk("atmio-E: RT irq not available\n");
+			return -EINVAL;
+		}
+		devpriv->rt_irq=dev->irq;
+		devpriv->irqmask=mask;
+	}else{
+		devpriv->irqmask|=mask;
+	}
+	ni_rt_interrupt_dev=dev;
+
+	return 0;
+}
+
+static int ungrab_rt_irq(comedi_device *dev,int mask)
+{
+	if(!(devpriv->irqmask&mask))return 0;
+
+	devpriv->irqmask&=~mask;
+	if(devpriv->irqmask)return 0;
+
+	if(devpriv->rt_irq){
+		rtl_free_global_irq(devpriv->rt_irq);
+		devpriv->rt_irq=0;
+	}else{
+		rt_printk("ni-E: RT irq already free\n");
+	}
+	return 0;
+}
+
+#if LINUX_VERSION_CODE < 0x020200
+static void ni_E_rt_interrupt(void)
 {
 	/* XXX this is terrible, but it's a limitation of RTLinux */
-	comedi_device *dev=ni_rt_interrupt_dev;
+	do_interrupt(ni_rt_interrupt_dev);
+}
 #else
+static unsigned int ni_E_rt_interrupt(unsigned int irq,struct pt_regs *regs)
+{
+	/* XXX this is terrible, but it's a limitation of RTLinux */
+	do_interrupt(ni_rt_interrupt_dev);
+
+	return 0;
+}
+#endif
+
+#else /* !CONFIG_COMEDI_RT */
+#define grab_rt_irq(a,b)
+#define ungrab_rt_irq(a,b)
+#endif
+
 static void ni_E_interrupt(int irq,void *d,struct pt_regs * regs)
 {
-	comedi_device *dev=d;
-#endif
+	do_interrupt(d);
+}
+
+static void do_interrupt(comedi_device *dev)
+{
 #ifdef TIMEMARK
 	struct timeval tv;
 #endif
 	comedi_subdevice *s=dev->subdevices;
 	int status;
+	unsigned short b_status;
 	int ack=0;
 	int wsave;
 
@@ -172,65 +252,104 @@ static void ni_E_interrupt(int irq,void *d,struct pt_regs * regs)
 */
 	wsave=win_save();
 	
+	b_status=ni_readw(AO_Status_1);
+#if 0
+printk("B status=0x%04x\n",b_status);
+#endif
 	status=ni_readw(AI_Status_1);
-	if(status&(AI_Overrun_St|AI_Overflow_St))
-		printk("ni_E: overrun/overflow status=0x%04x\n",status);
-	
+	if(status&(AI_Overrun_St|AI_Overflow_St)){
+		rt_printk("ni_E: overrun/overflow status=0x%04x\n",status);
+		comedi_done(dev,s);
+		ungrab_rt_irq(dev,RTirqmask_AI);
+		return;
+	}
 
 #ifdef TIMEMARK
 do_gettimeofday(&tv);
-printk("ni-E %d.%06d status=0x%04x\n",tv.tv_sec,tv.tv_usec,status);
+rt_printk("ni-E %d.%06d status=0x%04x\n",tv.tv_sec,tv.tv_usec,status);
 #endif
 
-if(!dev->subdevices->cur_trig.data){
-       printk("aiee! cur_trig.data got zapped!\n");
-       comedi_done(dev,s);
-       return;
-}
+#if 0
+	/* this is wrong, since we do AO now  */
+	if(!dev->subdevices->cur_trig.data){
+		rt_printk("aiee! cur_trig.data got zapped!\n");
+		comedi_done(dev,s);
+		ungrab_rt_irq(dev,RTirqmask_AI);
+		return;
+	}
+#endif
 
 	if(status&AI_SC_TC_St){
 #ifdef DEBUG
-printk("ni-E: SC_TC interrupt\n");
+rt_printk("ni-E: SC_TC interrupt\n");
 #endif
 		ni_handle_fifo_dregs(dev);
 		win_out(0x0000,Interrupt_A_Enable_Register);
 		comedi_done(dev,s);
+		ungrab_rt_irq(dev,RTirqmask_AI);
 
 		ack|=AI_SC_TC_Interrupt_Ack;
 	}
 	if(status&AI_FIFO_Half_Full_St){
 		ni_handle_fifo_half_full(dev);
 	}
-	if((dev->subdevices->cur_trig.flags&TRIG_WAKE_EOS) && !(status&AI_FIFO_Empty_St)){
-		int new_cur_chan;
-
+	if(!(status&AI_FIFO_Empty_St)
+	   && ((dev->subdevices->cur_trig.flags&TRIG_WAKE_EOS)
+	      || (dev->subdevices->cb_mask&COMEDI_CB_EOS))){
 		ni_handle_fifo_dregs(dev);
 
-		new_cur_chan=s->buf_int_ptr/s->cur_trig.n_chan;
-		if(new_cur_chan>s->cur_chan){
-#ifdef DEBUG
-printk("buf_int_ptr=%d\n",s->buf_int_ptr);
-#endif
-			s->cur_chan=new_cur_chan;
+		if(s->buf_int_count>=s->cur_chan){
+			while(s->buf_int_count>=s->cur_chan)
+				s->cur_chan+=s->cur_trig.n_chan*sizeof(sampl_t);
 			comedi_eos(dev,dev->subdevices+0);
 		}
 	}
 
+	if(b_status&AO_Overrun_St){
+		printk("ni-E: AO FIFO underrun status=0x%04x status2=0x%04x\n",b_status,ni_readw(AO_Status_2));
+	}
+
+	if(b_status&AO_BC_TC_St){
+		printk("ni-E: AO BC_TC status=0x%04x status2=0x%04x\n",b_status,ni_readw(AO_Status_2));
+	}
+
+	if(b_status&AO_FIFO_Request_St)
+		ni_ao_fifo_half_empty(dev,dev->subdevices+1);
+
+	b_status=ni_readw(AO_Status_1);
+	if(b_status&Interrupt_B_St){
+		if(b_status&AO_FIFO_Request_St){
+			printk("AO buffer underrun\n");
+		}
+		printk("Ack! didn't clear AO interrupt. b_status=0x%04x\n",b_status);
+		win_out(0,Interrupt_B_Enable_Register);
+	}
+
 	if(ack){
 		/* the AI_Configuration_Start is required */
 
 		win_out(AI_Configuration_Start,Joint_Reset_Register);
-		ni_writew(ack,Interrupt_A_Ack_Register);
+		ni_writew(ack,Interrupt_A_Ack_Register); /* XXX "_Register" is wrongo... */
 		win_out(AI_Configuration_End,Joint_Reset_Register);
 	}
 
 	win_restore(wsave);
 }
 
+static void ni_ai_fifo_read(comedi_device *dev,comedi_subdevice *s,
+		sampl_t *data,int n)
+{
+	int i;
+
+	for(i=0;i<n;i++){
+		data[i]=ni_readw(ADC_FIFO_Data_Register);
+	}
+}
+
 
 static void ni_handle_fifo_half_full(comedi_device *dev)
 {
-	int i;
+	int n,m;
 	comedi_subdevice *s=dev->subdevices+0;
 
 	/*
@@ -243,30 +362,25 @@ static void ni_handle_fifo_half_full(comedi_device *dev)
 
 	*/
 	
-if(s->buf_int_ptr+boardtype.ai_fifo_depth/2 > s->buf_len){
-       printk("comedi: ni-E: (bug) about to overrun buffer HF\n");
-       s->buf_int_ptr=0;
-}
+	n=boardtype.ai_fifo_depth/2;
 
-	/*
-	   This will warn us if the FIFO depth is documented wrong.
-	 */
-	for(i=0;i<boardtype.ai_fifo_depth/2-1;i++)
-		s->cur_trig.data[s->buf_int_ptr++]=ni_readw(ADC_FIFO_Data_Register);
-	if(ni_readw(AI_Status_1)&AI_FIFO_Empty_St)
-		printk("comedi: ni-E: (bug) FIFO smaller than expected\n");
-	s->cur_trig.data[s->buf_int_ptr++]=ni_readw(ADC_FIFO_Data_Register);
+	/* this makes the assumption that the buffer length is
+	   greater than the half-fifo depth. */
 
-#if 0
-	/* debug code to determine actual fifo depth */
-	
-	while(!(ni_readw(AI_Status_1)&AI_FIFO_Empty_St)){
-		s->cur_trig.data[s->buf_int_ptr++]=ni_readw(ADC_FIFO_Data_Register);
-		i++;
+	if(s->buf_int_ptr+n*sizeof(sampl_t)>=s->cur_trig.data_len){
+		m=(s->cur_trig.data_len-s->buf_int_ptr)/sizeof(sampl_t);
+		ni_ai_fifo_read(dev,s,((void *)(s->cur_trig.data))+s->buf_int_ptr,m);
+		s->buf_int_count+=m*sizeof(sampl_t);
+		n-=m;
+		s->buf_int_ptr=0;
+
+		comedi_eobuf(dev,s);
 	}
-	printk("comedi: ni-E: apparent FIFO half-depth %d\n",i);
-#endif
+	ni_ai_fifo_read(dev,s,((void *)(s->cur_trig.data))+s->buf_int_ptr,n);
+	s->buf_int_count+=n*sizeof(sampl_t);
+	s->buf_int_ptr+=n*sizeof(sampl_t);
 
+	comedi_bufcheck(dev,s);
 }
 
 /*
@@ -275,6 +389,8 @@ if(s->buf_int_ptr+boardtype.ai_fifo_depth/2 > s->buf_len){
 static void ni_handle_fifo_dregs(comedi_device *dev)
 {
 	comedi_subdevice *s=dev->subdevices+0;
+	sampl_t *data;
+	int i,n;
 
 	/*
 	   Too bad NI didn't have the foresight to return a
@@ -286,20 +402,21 @@ static void ni_handle_fifo_dregs(comedi_device *dev)
 	   This would save a lot of time.
 	*/
 
-	while(!(ni_readw(AI_Status_1)&AI_FIFO_Empty_St)){
-if(s->buf_int_ptr >= s->buf_len){
-	printk("comedi: ni-E: buffer overrun HF\n");
-	ni_readw(ADC_FIFO_Data_Register);
-}else{
-		s->cur_trig.data[s->buf_int_ptr++]=ni_readw(ADC_FIFO_Data_Register);
-}
+	data=((void *)s->cur_trig.data)+s->buf_int_ptr;
+	while(1){
+		n=(s->cur_trig.data_len-s->buf_int_ptr)/sizeof(sampl_t);
+		for(i=0;i<n;i++){
+			if(ni_readw(AI_Status_1)&AI_FIFO_Empty_St)
+				return;
+			*data=ni_readw(ADC_FIFO_Data_Register);
+			data++;
+			s->buf_int_ptr+=sizeof(sampl_t);
+			s->buf_int_count+=sizeof(sampl_t);
+		}
+		s->buf_int_ptr=0;
+		data=s->cur_trig.data;
+		comedi_eobuf(dev,s);
 	}
-#if 0
-	while(!(ni_readw(AI_Status_1)&AI_FIFO_Empty_St)){
-		ni_readw(ADC_FIFO_Data_Register);
-		s->buf_int_ptr++;
-	}
-#endif
 }
 
 /*
@@ -309,6 +426,8 @@ if(s->buf_int_ptr >= s->buf_len){
  */
 static int ni_ai_reset(comedi_device *dev,comedi_subdevice *s)
 {
+	ungrab_rt_irq(dev,RTirqmask_AI);
+
 	win_out(0x0000,Interrupt_A_Enable_Register);
 
 	win_out(AI_Reset,Joint_Reset_Register);
@@ -359,6 +478,7 @@ static void ni_load_channelgain_list(comedi_device *dev,unsigned int n_chan,unsi
 static int ni_ai_mode0(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 {
 	int i;
+	int chan;
 	int wsave;
 	
 	wsave=win_save();
@@ -368,13 +488,8 @@ static int ni_ai_mode0(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 	/* interrupt on errors */
 	win_out(0x0020,Interrupt_A_Enable_Register) ;
 	
-#if 0
-	/* this function needs to be fixed for multiple channels */
-	ni_load_channelgain_list(dev,it->n_chan,it->chanlist);
-#else
-	ni_load_channelgain_list(dev,1,it->chanlist);
-#endif
-
+	for(chan=0;chan<it->n_chan;chan++){
+		ni_load_channelgain_list(dev,1,it->chanlist+chan);
 #if 0
 	/* needs start configuration */
 	win_out(AI_START_Edge|AI_START_Sync|
@@ -383,26 +498,34 @@ static int ni_ai_mode0(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 #endif
 
 
-	win_out(1,AI_Command_1_Register);
+		win_out(1,AI_Command_1_Register);
 	
-	for(i=0;i<40;i++){
-		if(!(ni_readw(AI_Status_1)&AI_FIFO_Empty_St)){
-			it->data[0]=ni_readw(ADC_FIFO_Data_Register);
-			if(boardtype.adbits==12){
-				if(CR_RANGE(it->chanlist[0])<8)
-					it->data[0]^=0x800;
-				it->data[0]&=0xfff;
-			}else{
-				if(CR_RANGE(it->chanlist[0])<8)
-					it->data[0]^=0x8000;
-				it->data[0]&=0xffff;
+		/* I don't know how long it takes to access the bus,
+		   so shorter loops might cause timeouts */
+#define NI_TIMEOUT 1000
+		for(i=0;i<NI_TIMEOUT;i++){
+			if(!(ni_readw(AI_Status_1)&AI_FIFO_Empty_St)){
+				it->data[chan]=ni_readw(ADC_FIFO_Data_Register);
+				if(boardtype.adbits==12){
+					if(CR_RANGE(it->chanlist[chan])<8)
+						it->data[chan]^=0x800;
+					it->data[chan]&=0xfff;
+				}else{
+					if(CR_RANGE(it->chanlist[chan])<8)
+						it->data[chan]^=0x8000;
+					it->data[chan]&=0xffff;
+				}
+				break;
 			}
-			win_restore(wsave);
-			return 1;
+			/*udelay(25);*/
 		}
-		udelay(25);
+		if(i==NI_TIMEOUT)goto timeout;
 	}
-	printk("ni_E: timeout 2\n");
+	win_restore(wsave);
+	return chan;
+
+timeout:
+	rt_printk("ni_E: timeout 2\n");
 	win_restore(wsave);
 	return -ETIME;
 }
@@ -536,7 +659,6 @@ static void ni_load_channelgain_list(comedi_device *dev,unsigned int n_chan,unsi
 
 		lo=((i==n_chan-1)?0x8000:0) | ((range&0x8)<<5) | (range&0x7);
 		ni_writew(lo,Configuration_Memory_Low);
-DPRINTK("%x %x\n",hi,lo);
 	}
 
 	/* prime the channel/gain list */
@@ -549,7 +671,7 @@ DPRINTK("%x %x\n",hi,lo);
 		}
 		udelay(25);
 	}
-	printk("ni_E: timeout 1\n");
+	rt_printk("ni_E: timeout 1\n");
 }
 
 
@@ -560,6 +682,9 @@ static int ni_ai_mode2(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 {
 	int wsave;
 
+	if(it->flags&TRIG_RT)
+		grab_rt_irq(dev,RTirqmask_AI);
+
 	wsave=win_save();
 
 	win_out(1,ADC_FIFO_Clear);
@@ -625,7 +750,7 @@ static int ni_ai_mode2(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 		/* interrupt on FIFO, errors, SC_TC */
 		bits=0x00a1;
 
-		if(it->flags&TRIG_WAKE_EOS){
+		if(it->flags&TRIG_WAKE_EOS || s->cb_mask&COMEDI_CB_EOS){
 			/*generate FIFO interrupts on non-empty */
 			win_out(AI_FIFO_Mode_NE|0x0000,AI_Mode_3_Register);
 #if 0
@@ -646,7 +771,7 @@ static int ni_ai_mode2(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 	}
 
 #ifdef DEBUG
-printk("end config\n");
+rt_printk("end config\n");
 #endif
 	/* end configuration */
 	win_out(AI_Configuration_End,Joint_Reset_Register);
@@ -657,11 +782,11 @@ printk("end config\n");
 	/* AI_START1_Pulse */
 	win_out(AI_START1_Pulse,AI_Command_2_Register);
 #ifdef DEBUG
-printk("START1 pulse\n");
+rt_printk("START1 pulse\n");
 #endif
 
 	/* XXX hack alert */
-	s->cur_chan=0;
+	s->cur_chan=s->cur_trig.n_chan*sizeof(sampl_t);
 
 	win_restore(wsave);
 	return 0;
@@ -675,6 +800,9 @@ static int ni_ai_mode4(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 {
 	int wsave;
 
+	if(it->flags&TRIG_RT)
+		grab_rt_irq(dev,RTirqmask_AI);
+
 	wsave=win_save();
 
 	win_out(1,ADC_FIFO_Clear);
@@ -742,7 +870,7 @@ static int ni_ai_mode4(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 		/* interrupt on FIFO, errors, SC_TC */
 		bits=0x00a1;
 
-		if(it->flags&TRIG_WAKE_EOS){
+		if(it->flags&TRIG_WAKE_EOS || s->cb_mask&COMEDI_CB_EOS){
 			/* generate FIFO interrupts on non-empty */
 			win_out(AI_SI2_Source_Select|AI_FIFO_Mode_NE|0x0000,AI_Mode_3_Register);
 		}else{
@@ -771,31 +899,215 @@ static int ni_ai_mode4(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 	win_out(AI_START1_Pulse,AI_Command_2_Register);
 
 	/* XXX hack alert */
-	s->cur_chan=0;
+	s->cur_chan=s->cur_trig.n_chan*sizeof(sampl_t);
 
 	win_restore(wsave);
 	return 0;
 }
 
 
-static int ni_ao_mode0(comedi_device *dev,comedi_subdevice *s,comedi_trig *it);
-
-#if 0
-/*
-	analog output switch
-*/
-static int ni_ao(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
+static void ni_ao_fifo_load(comedi_device *dev,comedi_subdevice *s,
+		sampl_t *data,int n)
 {
-	switch(it->mode){
-	case 0:
-		return ni_ao_mode0(dev,s,it);
-	default:
-		return -EINVAL;
+	int i;
+
+	for(i=0;i<n;i++){
+		ni_writew(data[i],DAC_FIFO_Data);
 	}
 }
+
+
+/*
+ *  There's a small problem if the FIFO gets really low and we
+ *  don't have the data to fill it.  Basically, if after we fill
+ *  the FIFO with all the data available, the FIFO is _still_
+ *  less than half full, we never clear the interrupt.  If the
+ *  IRQ is in edge mode, we never get another interrupt, because
+ *  this one wasn't cleared.  If in level mode, we get flooded
+ *  with interrupts that we can't fulfill, because nothing ever
+ *  gets put into the buffer.
+ *
+ *  This kind of situation is recoverable, but it is easier to
+ *  just pretend we had a FIFO underrun, since there is a good
+ *  chance it will happen anyway.  This is _not_ the case for
+ *  RT code, as RT code might purposely be running close to the
+ *  metal.  Needs to be fixed eventually.
+ */
+static int ni_ao_fifo_half_empty(comedi_device *dev,comedi_subdevice *s)
+{
+	int n,m;
+
+	n=(s->buf_int_count-s->buf_user_count)/sizeof(sampl_t);
+	if(n==0)return 0;
+	if(n>boardtype.ao_fifo_depth/2)
+		n=boardtype.ao_fifo_depth/2;
+
+	if(s->buf_int_ptr+n*sizeof(sampl_t)>s->cur_trig.data_len){
+		m=(s->cur_trig.data_len-s->buf_int_ptr)/sizeof(sampl_t);
+		ni_ao_fifo_load(dev,s,((void *)(s->cur_trig.data))+s->buf_int_ptr,m);
+		s->buf_int_count+=m*sizeof(sampl_t);
+		s->buf_int_ptr=0;
+		n-=m;
+	}
+	ni_ao_fifo_load(dev,s,((void *)(s->cur_trig.data))+s->buf_int_ptr,n);
+	s->buf_int_count+=n*sizeof(sampl_t);
+	s->buf_int_ptr+=n*sizeof(sampl_t);
+
+	comedi_bufcheck(dev,s);
+
+	return 1;
+}
+
+static int ni_ao_prep_fifo(comedi_device *dev,comedi_subdevice *s)
+{
+	int n;
+
+	/* reset fifo */
+	win_out(0,DAC_FIFO_Clear);
+
+	/* load some data */
+	n=(s->buf_int_count-s->buf_user_count)/sizeof(sampl_t);
+	if(n==0)return 0;
+	if(n>boardtype.ao_fifo_depth)
+		n=boardtype.ao_fifo_depth;
+
+	ni_ao_fifo_load(dev,s,((void *)(s->cur_trig.data))+s->buf_int_ptr,n);
+	s->buf_int_count+=n*sizeof(sampl_t);
+	s->buf_int_ptr+=n*sizeof(sampl_t);
+
+	return n;
+}
+
+
+#if 0
+static void AO_Reset_All(comedi_device *dev)
+{
+	win_out(AO_Configuration_Start,Joint_Reset_Register);
+
+	win_out(AO_Disarm,AO_Command_1_Register);
+	devpriv->cmd1=0;
+
+	win_out(0,Interrupt_B_Enable_Register);
+
+	win_out(AO_BC_Source_Select(1),AO_Personal_Register);
+
+	win_out(0x3f98,Interrupt_B_Ack_Register);
+
+	win_out(AO_Configuration_End,Joint_Reset_Register);
+}
+
+static void AO_Board_Personalize(comedi_device *dev)
+{
+	win_out(AO_Configuration_Start,Joint_Reset_Register);
+
+	win_out(0x1430,AO_Personal_Register);
+
+	win_out(0x1b20,Clock_and_FOUT_Register);
+
+	win_out(0,AO_Output_Control_Register);
+
+	win_out(0,AO_START_Select_Register);
+
+	win_out(AO_Configuration_End,Joint_Reset_Register);
+}
+
+static void AO_LDAC_Source_And_Update_Mode(comedi_device *dev)
+{
+	win_out(AO_Configuration_Start,Joint_Reset_Register);
+
+	devpriv->ao_cmd1&=~(AO_LDAC0_Source_Select|AO_DAC0_Update_Mode|AO_LDAC1_Source_Select|AO_DAC1_Update_Mode);
+	win_out(devpriv->ao_cmd1,AO_Command_1_Register);
+
+	win_out(AO_Configuration_End,Joint_Reset_Register);
+}
+
+static void AO_Triggering(comedi_device *dev)
+{
+	win_out(AO_Configuration_Start,Joint_Reset_Register);
+
+	devpriv->ao_mode1|=AO_Trigger_Once;
+	win_out(devpriv->ao_mode1,AO_Mode_1_Register);
+
+	devpriv->ao_trigger_select&=~(AO_START1_Select(-1)|AO_START1_Polarity);
+	devpriv->ao_trigger_select|=AO_START1_Edge|AO_START1_Sync;
+	win_out(devpriv->ao_trigger_select,AO_Trigger_Select_Register);
+
+	devpriv->ao_mode3&=~AO_Trigger_Length;
+	win_out(devpriv->ao_mode3,AO_Mode_3_Register);
+	
+	win_out(AO_Configuration_End,Joint_Reset_Register);
+}
+
+static void AO_Counting(comedi_device *dev)
+{
+	win_out(AO_Configuration_Start,Joint_Reset_Register);
+
+	devpriv->ao_mode1&=~AO_Continuous;
+	win_out(devpriv->ao_mode1,AO_Mode_1_Register);
+
+	devpriv->ao_mode2&=~AO_BC_Initial_Load_Source;
+	win_out(devpriv->ao_mode2,AO_Mode_2_Register);
+
+	win_out(0,AO_BC_Load_A_Register_High);
+	win_out(4,AO_BC_Load_A_Register_Low);
+	win_out(devpriv->ao_cmd1|AO_BC_Load,AO_Command_1_Register);
+
+	devpriv->ao_mode2&=~AO_UC_Initial_Load_Source;
+	win_out(devpriv->ao_mode2,AO_Mode_2_Register);
+	
+	win_out(0,AO_UC_Load_A_Register_High);
+	win_out(3000,AO_UC_Load_A_Register_Low);
+	win_out(devpriv->ao_cmd1|AO_UC_Load,AO_Command_1_Register);
+
+	win_out(0,AO_UC_Load_A_Register_High);
+	win_out(2999,AO_UC_Load_A_Register_Low);
+
+	win_out(AO_Configuration_End,Joint_Reset_Register);
+}
+
+static void AO_Updating(comedi_device *dev)
+{
+	win_out(AO_Configuration_Start,Joint_Reset_Register);
+
+	devpriv->ao_cmd2&=~AO_BC_Gate_Enable;
+	win_out(devpriv->ao_cmd2,AO_Command_2_Register);
+
+	devpriv->ao_mode1&=~(AO_UPDATE_Source_Select(-1)|AO_UPDATE_Source_Polarity);
+	devpriv->ao_mode1&=~(AO_UI_Source_Select(-1)|AO_UI_Source_Polarity);
+	win_out(devpriv->ao_mode1,AO_Mode_1_Register);
+
+	devpriv->ao_mode2&=~(AO_UI_Initial_Load_Source|AO_UI_Reload_Mode);
+	win_out(devpriv->ao_mode2,AO_Mode_2_Register);
+
+	win_out(0,AO_UI_Load_A_Register_High);
+	win_out(1,AO_UI_Load_A_Register_Low);
+	win_out(devpriv->ao_cmd1|AO_UI_Load,AO_Command_1_Register);
+
+	win_out(0,AO_UI_Load_A_Register_High);
+	win_out(0x9c40,AO_UI_Load_A_Register_Low);
+
+	win_out(AO_Configuration_End,Joint_Reset_Register);
+}
+
+static void AO_Channels(comedi_device *dev)
+{
+	win_out(AO_Configuration_Start,Joint_Reset_Register);
+
+	devpriv->ao_mode1&=~AO_Multiple_Channels;
+	win_out(devpriv->ao_mode1,AO_Mode_1_Register);
+
+	win_out(AO_Number_Of_Channels(1),AO_Output_Control_Register);
+
+	win_out(AO_Configuration_End,Joint_Reset_Register);
+}
+
+
+
 #endif
 
 
+
+
 static int ni_ao_mode0(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 {
 	unsigned int data;
@@ -832,6 +1144,128 @@ static int ni_ao_mode0(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 	return 1;
 }
 
+static int ni_ao_mode2(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
+{
+	unsigned int conf;
+	unsigned int chan;
+	unsigned int range;
+	
+	chan=CR_CHAN(it->chanlist[0]);
+
+	conf=chan<<8;
+	
+	/* XXX check range with current range in flaglist[chan] */
+	/* should update calibration if range changes (ick) */
+
+	range = CR_RANGE(it->chanlist[0]);
+	conf |= (range&1);
+	conf |= (range&2)<<1;
+	
+	/* not all boards can deglitch, but this shouldn't hurt */
+	if(it->flags & TRIG_DEGLITCH)
+		conf |= 2;
+
+	/* analog reference */
+	/* AREF_OTHER connects AO ground to AI ground, i think */
+	conf |= (CR_AREF(it->chanlist[0])==AREF_OTHER)? 8 : 0;
+
+	win_out(AO_Disarm,AO_Command_1_Register);
+
+	win_out(conf,AO_Configuration);
+
+	/* user is supposed to write() to buffer before triggering */
+	if(ni_ao_prep_fifo(dev,s)==0)
+		return -EIO;
+
+	win_out(AO_Configuration_Start,Joint_Reset_Register);
+
+	devpriv->ao_mode1|=AO_Trigger_Once;
+	win_out(devpriv->ao_mode1,AO_Mode_1_Register);
+	devpriv->ao_trigger_select&=~(AO_START1_Polarity|AO_START1_Select(-1));
+	devpriv->ao_trigger_select|=AO_START1_Edge|AO_START1_Sync;
+	win_out(devpriv->ao_trigger_select,AO_Trigger_Select_Register);
+	devpriv->ao_mode3&=~AO_Trigger_Length;
+	win_out(devpriv->ao_mode3,AO_Mode_3_Register);
+
+	if(it->n==0){
+		devpriv->ao_mode1|=AO_Continuous;
+printk("setting continuous\n");
+	}else{
+		devpriv->ao_mode1&=~AO_Continuous;
+printk("unsetting continuous\n");
+	}
+	win_out(devpriv->ao_mode1,AO_Mode_1_Register);
+	devpriv->ao_mode2&=~AO_BC_Initial_Load_Source;
+	win_out(devpriv->ao_mode2,AO_Mode_2_Register);
+	if(it->n==0){
+		win_out(0xff,AO_BC_Load_A_Register_High);
+		win_out(0xffff,AO_BC_Load_A_Register_Low);
+	}else{
+		win_out(0,AO_BC_Load_A_Register_High);
+		win_out(0,AO_BC_Load_A_Register_Low);
+	}
+	win_out(AO_BC_Load,AO_Command_1_Register);
+	devpriv->ao_mode2&=~AO_UC_Initial_Load_Source;
+	win_out(devpriv->ao_mode2,AO_Mode_2_Register);
+	if(it->n==0){
+		win_out(0xff,AO_UC_Load_A_Register_High);
+		win_out(0xffff,AO_UC_Load_A_Register_Low);
+		win_out(AO_UC_Load,AO_Command_1_Register);
+		win_out(0xff,AO_UC_Load_A_Register_High);
+		win_out(0xffff,AO_UC_Load_A_Register_Low);
+	}else{
+		win_out(0,AO_UC_Load_A_Register_High);
+		win_out(0,AO_UC_Load_A_Register_Low);
+		win_out(AO_UC_Load,AO_Command_1_Register);
+		win_out((it->n-1)>>16,AO_UC_Load_A_Register_High);
+		win_out((it->n-1)&0xffff,AO_UC_Load_A_Register_Low);
+	}
+
+	devpriv->ao_cmd2&=~AO_BC_Gate_Enable;
+	ni_writew(devpriv->ao_cmd2,AO_Command_2);
+	devpriv->ao_mode1&=~(AO_UI_Source_Select(0x1f)|AO_UI_Source_Polarity);
+	win_out(devpriv->ao_mode1,AO_Mode_1_Register);
+	devpriv->ao_mode2&=~(AO_UI_Reload_Mode(3)|AO_UI_Initial_Load_Source);
+	win_out(devpriv->ao_mode2,AO_Mode_2_Register);
+	win_out(0,AO_UI_Load_A_Register_High);
+	win_out(1,AO_UI_Load_A_Register_Low);
+	win_out(AO_UI_Load,AO_Command_1_Register);
+	win_out((it->trigvar>>16),AO_UI_Load_A_Register_High);
+	win_out((it->trigvar&0xffff),AO_UI_Load_A_Register_Low);
+
+	devpriv->ao_mode1&=~AO_Multiple_Channels;
+	win_out(devpriv->ao_mode1,AO_Mode_1_Register);
+	win_out(AO_UPDATE_Output_Select(1),AO_Output_Control_Register);
+
+	win_out(AO_DAC0_Update_Mode|AO_DAC1_Update_Mode,AO_Command_1_Register);
+
+	devpriv->ao_mode3|=AO_Stop_On_Overrun_Error;
+	win_out(devpriv->ao_mode3,AO_Mode_3_Register);
+
+devpriv->ao_mode2|=AO_FIFO_Mode(1);
+	devpriv->ao_mode2&=~AO_FIFO_Retransmit_Enable;
+	win_out(devpriv->ao_mode2,AO_Mode_2_Register);
+
+	win_out(AO_Configuration_End,Joint_Reset_Register);
+
+	win_out(devpriv->ao_mode3|AO_Not_An_UPDATE,AO_Mode_3_Register);
+	win_out(devpriv->ao_mode3,AO_Mode_3_Register);
+
+	/* wait for DACs to be loaded */
+	udelay(100);
+
+	win_out(devpriv->ao_cmd1|AO_UI_Arm|AO_UC_Arm|AO_BC_Arm|AO_DAC1_Update_Mode|AO_DAC0_Update_Mode,
+		AO_Command_1_Register);
+
+	win_out(AO_FIFO_Interrupt_Enable|AO_Error_Interrupt_Enable,Interrupt_B_Enable_Register);
+
+
+	ni_writew(devpriv->ao_cmd2|AO_START1_Pulse,AO_Command_2);
+	
+	return 0;
+}
+
+
 
 /*
 	digital i/o
@@ -840,7 +1274,7 @@ static int ni_ao_mode0(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 */
 static int ni_dio(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 {
-	int data,mask,data_in;
+	int data,mask;
 	int i;
 	int temp;
 
@@ -858,28 +1292,18 @@ static int ni_dio(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 		s->io_bits=data;
 		win_out(s->io_bits,DIO_Control_Register);
 	}else{
-		data=devpriv->last_do;
-		data_in=win_in(DIO_Input_Register);
-		for(i=0;i<it->n_chan;i++){
-			mask=1<<CR_CHAN(it->chanlist[i]);
-			if(s->io_bits & mask){
-				/* output bit */
-				data&= ~mask;
-				if(it->data[i])
-					data|=mask;
-			}else{
-				/* input bit */
-				it->data[i]=(data_in&mask)?1:0;
-			}
+		if(it->flags & TRIG_WRITE){
+			do_pack(&s->state,it);
+			win_out(s->state,DIO_Output_Register);
+		}else{
+			data=win_in(DIO_Input_Register);
+			di_unpack(data,it);
 		}
-		devpriv->last_do=data;
-
-		win_out(data,DIO_Output_Register);
 	}
 
 	win_restore(temp);
 
-	return i;
+	return it->n_chan;
 }
 
 #if 0
@@ -967,11 +1391,8 @@ static int ni_E_init(comedi_device *dev,comedi_devconfig *it)
 	
 	dev->n_subdevices=7;
 	
-	dev->subdevices=kmalloc(dev->n_subdevices*sizeof(comedi_subdevice),GFP_KERNEL);
-	if(!dev->subdevices){
+	if(alloc_subdevices(dev)<0)
 		return -ENOMEM;
-	}
-	memset(dev->subdevices,0,dev->n_subdevices*sizeof(comedi_subdevice));
 	
 	/* analog input subdevice */
 
@@ -996,13 +1417,18 @@ static int ni_E_init(comedi_device *dev,comedi_devconfig *it)
 	/* XXX what about boards without ao? */
 
 	s=dev->subdevices+1;
-	s->type=COMEDI_SUBD_AO;
-	s->subdev_flags=SDF_WRITEABLE|SDF_RT|SDF_DEGLITCH|SDF_GROUND|SDF_OTHER;
-	s->n_chan=2;
-	s->maxdata=(1<<12)-1;
-	s->timer_type=TIMER_atmio;
-	s->range_type=RANGE_ni_E_ao_ext;	/* XXX wrong for some boards */
-	s->trig[0]=ni_ao_mode0;
+	if(boardtype.n_aochan){
+		s->type=COMEDI_SUBD_AO;
+		s->subdev_flags=SDF_WRITEABLE|SDF_RT|SDF_DEGLITCH|SDF_GROUND|SDF_OTHER;
+		s->n_chan=boardtype.n_aochan;
+		s->maxdata=(1<<boardtype.aobits)-1;
+		s->timer_type=TIMER_atmio;
+		s->range_type=RANGE_ni_E_ao_ext;	/* XXX wrong for some boards */
+		s->trig[0]=ni_ao_mode0;
+		s->trig[2]=ni_ao_mode2;
+	}else{
+		s->type=COMEDI_SUBD_UNUSED;
+	}
 	
 	/* digital i/o subdevice */
 	
@@ -1038,7 +1464,7 @@ static int ni_E_init(comedi_device *dev,comedi_devconfig *it)
 	s=dev->subdevices+5;
 	s->type=COMEDI_SUBD_CALIB;
 	s->subdev_flags=SDF_WRITEABLE|SDF_INTERNAL;
-	s->n_chan=(boardtype.caldactype==1)?12:9;
+	s->n_chan=(boardtype.caldactype==caldac_mb88341)?12:9;
 	s->maxdata=0xfff;
 	s->trig[0]=ni_calib;
 	
@@ -1062,20 +1488,28 @@ static int ni_E_init(comedi_device *dev,comedi_devconfig *it)
 	devpriv->ao1p=0x0100;
 	ni_writew(devpriv->ao1p,AO_Configuration);
 	win_out(AO_Configuration_Start,Joint_Reset_Register);
-	win_out(0x2000,AO_Command_1_Register);
+	win_out(AO_Disarm,AO_Command_1_Register);
 	win_out(0,Interrupt_B_Enable_Register);
 	win_out(0x0010,AO_Personal_Register);
 	ni_writew(0x3f98,Interrupt_B_Ack);
 	win_out(0x1430,AO_Personal_Register);
 	win_out(0,AO_Output_Control_Register);
 	win_out(0,AO_Start_Select_Register);
-	win_out(0,AO_Command_1_Register);
+	devpriv->ao_cmd1=0;
+	win_out(devpriv->ao_cmd1,AO_Command_1_Register);
+	devpriv->ao_cmd2=0;
+	devpriv->ao_mode1=0;
+	devpriv->ao_mode2=0;
+	devpriv->ao_mode3=0;
+	devpriv->ao_trigger_select=0;
 
         if(dev->irq){
                 win_out((IRQ_POLARITY<<0) |  /* polarity : active high */
                         (0<<1) |  /* no interrupt on 3 pins */
                         (1<<11) |  /* enable int A */
-                        (interrupt_pin(dev->irq)<<8),  /* interrupt output pin */
+                        (1<<15) |  /* enable int B */
+                        (interrupt_pin(dev->irq)<<8) |  /* interrupt output pin A */
+                        (interrupt_pin(dev->irq)<<12) ,  /* interrupt output pin B */
                         Interrupt_Control_Register
                 );
         }
@@ -1108,12 +1542,12 @@ static int ni_rem(comedi_device *dev)
 
 static int ni_eeprom(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 {
-	if(it->mode != 0)
-		return -EINVAL;
+	int i;
+	
+	for(i=0;i<it->n_chan;i++)
+		it->data[i]=ni_read_eeprom(dev,CR_CHAN(it->chanlist[i]));
 
-	it->data[0]=ni_read_eeprom(dev,CR_CHAN(it->chanlist[0]));
-
-	return 1;
+	return i;
 }
 
 /*
@@ -1148,9 +1582,6 @@ static void ni_write_caldac(comedi_device *dev,int addr,int val);
 */
 static int ni_calib(comedi_device *dev,comedi_subdevice *s,comedi_trig *it)
 {
-	if(it->mode != 0)
-		return -EINVAL;
-
 	ni_write_caldac(dev,CR_CHAN(it->chanlist[0]),it->data[0]);
 
 	return 1;
@@ -1161,7 +1592,7 @@ static void ni_write_caldac(comedi_device *dev,int addr,int val)
 	int loadbit,bits,bit,bitstring;
 	
 	switch(boardtype.caldactype){
-	case 1:
+	case caldac_mb88341:
 		/*
 		   MB 88341
 		   Note that address bits are reversed.  Thanks to
@@ -1180,7 +1611,7 @@ static void ni_write_caldac(comedi_device *dev,int addr,int val)
 			  (val&0xff);
 		loadbit=0x08;
 		break;
-	case 2:
+	case caldac_dac8800:
 		if(addr==8){
 			/*
 			   DAC 8043
