@@ -25,10 +25,11 @@
  */
 
 /*
-   Dave Andruczyk <xxx@tech.buffalostate.edu> also wrote a
+   Dave Andruczyk <dave@tech.buffalostate.edu> also wrote a
    driver for the PCL-711.  I used a few ideas from his driver
    here.  His driver also has more comments, if you are
    interested in understanding how this driver works.
+   http://tech.buffalostate.edu/~dave/driver/
 
    The ACL-8112 driver was hacked from the sources of the PCL-711
    driver (the 744 chip used on the 8112 is almost the same as
@@ -160,7 +161,6 @@ typedef struct {
 	int adchan;
 	int ntrig;
 	int aip[8];
-	int last_do;
 	int mode;
 } pcl711_private;
 
@@ -254,7 +254,7 @@ for(n=0;n<nmax;n++){
 			goto ok;
 		udelay(5);
 	}
-	printk("comedi%d: pcl711: A/D timeout\n", dev->minor);
+	rt_printk("comedi%d: pcl711: A/D timeout\n", dev->minor);
 	return -ETIME;
 	
 ok:
@@ -357,7 +357,7 @@ static int pcl711_do(comedi_device * dev, comedi_subdevice * s, comedi_trig * it
 	int chan;
 	int i;
 
-	data=devpriv->last_do;
+	data=s->state;
 	for(i=0;i<it->n_chan;i++){
 		chan=CR_CHAN(it->chanlist[i]);
 		mask=(1<<chan);
@@ -367,7 +367,7 @@ static int pcl711_do(comedi_device * dev, comedi_subdevice * s, comedi_trig * it
 	}
 	outb(data & 0xff, dev->iobase + PCL711_DO_LO);
 	outb((data >> 8), dev->iobase + PCL711_DO_HI);
-	devpriv->last_do = data;
+	s->state = data;
 
 	return it->n_chan;
 }
@@ -378,15 +378,8 @@ static void free_resources(comedi_device * dev)
 	if (dev->irq)
 		free_irq(dev->irq, dev);
 
-	if (dev->private)
-		kfree(dev->private);
-
 	if (dev->iobase)
 		release_region(dev->iobase, dev->iosize);
-
-	if (dev->subdevices)
-		kfree(dev->subdevices);
-
 }
 
 /*  Initialization */
@@ -446,17 +439,11 @@ int pcl711_init(comedi_device * dev, comedi_devconfig * it)
 	}
 	dev->irq = irq;
 
-	ret = -ENOMEM;
-	dev->private = kmalloc(sizeof(pcl711_private), GFP_KERNEL);
-
 	dev->n_subdevices = 4;
-	dev->subdevices = kmalloc(4 * sizeof(comedi_subdevice), GFP_KERNEL);
-
-	if (!dev->private || !dev->subdevices)
+	if((ret=alloc_subdevices(dev))<0)
+		goto fail;
+	if((ret=alloc_private(dev,sizeof(pcl711_private)))<0)
 		goto fail;
-
-	memset(dev->private, 0, sizeof(pcl711_private));
-	memset(dev->subdevices, 0, 4 * sizeof(comedi_subdevice));
 
 	s = dev->subdevices + 0;
 	/* AI subdevice */
@@ -499,6 +486,7 @@ int pcl711_init(comedi_device * dev, comedi_devconfig * it)
 	s->maxdata = 1;
 	s->len_chanlist = 16;
 	s->range_type = RANGE_digital;
+	s->state=0;
 	s->trig[0] = pcl711_do;
 
 	dev->rem = pcl711_rem;
