@@ -133,13 +133,15 @@ static int gaindelay[]={10,20,40,80};
 
 static int rti800_ai_mode0(comedi_device * dev, comedi_subdevice *s, comedi_trig * it)
 {
-	int i, hi, lo, gain;
+  int i;
+  for(i=0 ; i < it->n_chan ; i++) {
+    int t, hi, lo, gain;
 	int chan;
 	int data;
 	int status;
 
-	chan = CR_CHAN(it->chanlist[0]);
-	gain = CR_RANGE(it->chanlist[0]);
+    chan = CR_CHAN(it->chanlist[i]);
+    gain = CR_RANGE(it->chanlist[i]);
 
 	inb(dev->iobase + RTI800_ADCHI);
 	outb(0,dev->iobase+RTI800_CLRFLAGS);
@@ -150,23 +152,23 @@ static int rti800_ai_mode0(comedi_device * dev, comedi_subdevice *s, comedi_trig
 	udelay(gaindelay[gain]);
 
 	outb(0, dev->iobase + RTI800_CONVERT);
-	for (i = 0; i < RTI800_TIMEOUT; i++) {
+    for (t = 0; t < RTI800_TIMEOUT; t++) {
 		status=inb(dev->iobase+RTI800_CSR);
 #if DEBUG
-printk("status=%x\n",status);
+      rt_printk("status=%x\n",status);
 #endif
 		if(status & RTI800_OVERRUN){
-			printk("rti800: a/d overflow\n");
+			rt_printk("rti800: a/d overflow\n");
 			outb(0,dev->iobase+RTI800_CLRFLAGS);
 
 			return -ETIME;
 		}
 		if (status & RTI800_DONE)
 			break;
-		udelay(25);
+      udelay(8);
 	}
-	if(i==RTI800_CONVERT){
-		printk("rti800: timeout\n");
+    if(t==RTI800_TIMEOUT){
+		rt_printk("rti800: timeout\n");
 
 		return -ETIME;
 	}
@@ -176,12 +178,12 @@ printk("status=%x\n",status);
 	data = (hi << 8) | lo;
 	data &= 0xfff;
 	if (devpriv->adc_coding == adc_2comp) {
-printk("ack!\n");
 		data ^= 0x800;
 	}
-	it->data[0]=data;
+    it->data[i]=data;
+  }
+  return i;
 
-	return 1;
 }
 
 static int rti800_ai_mode1(comedi_device * dev, comedi_subdevice *s, comedi_trig * it)
@@ -192,11 +194,13 @@ static int rti800_ai_mode1(comedi_device * dev, comedi_subdevice *s, comedi_trig
 
 static int rti800_ao(comedi_device * dev, comedi_subdevice *s, comedi_trig * it)
 {
+  int i;
+  for(i=0 ; i < it->n_chan ; i++) {
 	int chan;
 	int data;
 
-	chan=CR_CHAN(it->chanlist[0]);
-	data=it->data[0];
+    chan=CR_CHAN(it->chanlist[i]);
+    data=it->data[i];
 
 	switch(chan){
 	case 0:
@@ -205,35 +209,38 @@ static int rti800_ao(comedi_device * dev, comedi_subdevice *s, comedi_trig * it)
 		}
 		outb(data & 0xff, dev->iobase + RTI800_DAC0LO);
 		outb(data >> 8, dev->iobase + RTI800_DAC0HI);
-
-		return 1;
+       break;
 	case 1:
 		if (devpriv->dac1_coding == dac_2comp) {
 			data ^= 0x800;
 		}
 		outb(data & 0xff, dev->iobase + RTI800_DAC1LO);
 		outb(data >> 8, dev->iobase + RTI800_DAC1HI);
-
-		return 1;
+       break;
 	default:
 		return -EINVAL;
 	}
+  }
+  return i;
 }
 
 static int rti800_di(comedi_device * dev, comedi_subdevice *s, comedi_trig * it)
 {
-	it->data[0] = inb(dev->iobase + RTI800_DI);
+	unsigned int bits;
+	
+	bits = inb(dev->iobase + RTI800_DI);
 
-	return 1;
+	return di_unpack(bits,it);
 }
 
 static int rti800_do(comedi_device * dev, comedi_subdevice *s, comedi_trig * it)
 {
-	int data=it->data[0];
+	do_pack(&s->state,it);
 
-	outb(data, dev->iobase + RTI800_DO);
+       /* Outputs are inverted... */
+       outb(s->state ^ 0xff, dev->iobase + RTI800_DO);
 
-	return 1;
+	return it->n_chan;
 }
 
 
@@ -423,12 +430,6 @@ static int rti800_rem(comedi_device * dev)
 {
 	printk("comedi%d: rti800: remove\n", dev->minor);
 
-	if(dev->private)
-		kfree(dev->private);
-
-	if(dev->subdevices)
-		kfree(dev->subdevices);
-
 	if(dev->iobase)
 		release_region(dev->iobase, dev->iosize);
 
