@@ -56,8 +56,10 @@ typedef struct {
 
 static int rti802_ao(comedi_device * dev, comedi_subdevice *s, comedi_trig * it)
 {
-	int chan = CR_CHAN(it->chanlist[0]);
-	int data = it->data[0];
+  int i;
+  for(i=0 ; i < it->n_chan ; i++) {
+    int chan = CR_CHAN(it->chanlist[i]);
+    int data = it->data[i];
 
 	if (devpriv->dac_coding[chan] == dac_2comp) {
 		data ^= 0x800;
@@ -65,8 +67,8 @@ static int rti802_ao(comedi_device * dev, comedi_subdevice *s, comedi_trig * it)
 	outb(chan, dev->iobase + RTI802_SELECT);
 	outb(data & 0xff, dev->iobase + RTI802_DATALOW);
 	outb(data >> 8, dev->iobase + RTI802_DATAHIGH);
-
-	return 0;
+  }
+  return i;
 }
 
 /*
@@ -101,16 +103,11 @@ int rti802_init(comedi_device * dev, comedi_devconfig * it)
 	dev->driver_name = "rti802";
 
 	dev->n_subdevices = 1;
-	dev->subdevices = kmalloc(sizeof(comedi_subdevice),GFP_KERNEL);
-	dev->private = kmalloc(sizeof(rti802_private) , GFP_KERNEL);
-	if (!dev->private || !dev->subdevices) {
+	if(alloc_subdevices(dev)<0 || alloc_private(dev,sizeof(rti802_private))){
 		rti802_free_resources(dev);
 		return -ENOMEM;
 	}
 
-	memset(dev->private, 0, sizeof(rti802_private));
-	memset(dev->subdevices, 0, sizeof(comedi_subdevice));
-
 	s=dev->subdevices;
 	/* ao subdevice */
 	s->type=COMEDI_SUBD_AO;
@@ -138,10 +135,6 @@ int rti802_init(comedi_device * dev, comedi_devconfig * it)
 
 static void rti802_free_resources(comedi_device * dev)
 {
-	if(dev->subdevices)
-		kfree(dev->subdevices);
-	if(dev->private)
-		kfree(dev->private);
 	if(dev->iobase)
 		release_region(dev->iobase, RTI802_SIZE);
 }
